# Copyright (C) 2009-2010, University of Maryland
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/ or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

# Author: Ismet Sahin
"""
BFGS quasi-newton optimizer.

All modules in this file are implemented from the book
"Numerical Methods for Unconstrained Optimization and Nonlinear Equations" by
J.E. Dennis and Robert B. Schnabel (Only a few minor modifications are done).

The interface is through the :func:`quasinewton` function.  Here is an
example call::

    n = 2
    x0 = [-0.9 0.9]'
    fn = lambda p: (1-p[0])**2 + 100*(p[1]-p[0]**2)**2
    grad = lambda p: array([-2*(1-p[0]) - 400*(p[1]-p[0]**2)*p[0], 200*p[1]])
    Sx = ones(n,1)
    typf = 1                       # todo. see what default value is the best
    macheps = eps
    eta = eps
    maxstep = 100
    gradtol = 1e-6
    steptol = 1e-12                # do not let steptol larger than 1e-9
    itnlimit = 1000
    result = quasinewton(fn, x0, grad, Sx, typf,
                         macheps, eta, maxstep, gradtola, steptol, itnlimit)
    print("status code %d"%result['status'])
    print("x_min=%s, f(x_min)=%g"%(str(result['x']),result['fx']))
    print("iterations, function calls, linesearch function calls",
          result['iterations'],result['evals'],result['linesearch_evals'])
"""

__all__ = ["quasinewton"]

from numpy import amax, array, diag, dot, finfo, inf, inner, isinf, isnan, linalg, maximum, ones, outer, sqrt, zeros

STATUS = {
    1: "Gradient < tolerance",
    2: "Step size < tolerance",
    3: "Invalid point in line search",
    4: "Iterations exceeded",
    5: "Max step taken --- function unbounded?",
    6: "User abort",
    7: "Iterations exceeded in line search",
    8: "Line search step size is too small",
    9: "Singular Hessian",
}


def quasinewton(
    fn,
    x0=None,
    grad=None,
    Sx=None,
    typf=1,
    macheps=None,
    eta=None,
    maxstep=100,
    gradtol=1e-6,
    steptol=1e-12,
    itnlimit=2000,
    abort_test=None,
    monitor=lambda **kw: True,
):
    r"""
    Run a quasinewton optimization on the problem.

    *fn(x)* is the cost function, which takes a point x and returns a scalar fx.

    *x0* is the initial point

    *grad* is the analytic gradient (if available)

    *Sx* is a scale vector indicating the typical values for parameters in
    the fitted result. This is used for a variety of things such as setting
    the step size in the finite difference approximation to the gradient, and
    controlling numerical accuracy in calculating the Hessian matrix.  If for
    example some of your model parameters are in the order of 1e-6, then Sx
    for those parameters should be set to 1e-6. Default: [1, ...]

    *typf* is the typical value for f(x) near the minimum.  This is used along
    with gradtol to check the gradient stopping condition.  Default: 1

    *macheps* is the minimum value that can be added to 1 to produce a number
    not equal to 1.  Default: numpy.finfo(float).eps

    *eta* adapts the numerical gradient calculations to machine precision.
    Default: *macheps*

    *maxstep* is the maximum step size in any gradient step, after normalizing
    by *Sx*. Default: 100

    *gradtol* is a stopping condition for the fit based on the amount of
    improvement expected at the next step.  Default: 1e-6

    *steptol* is a stopping condition for the fit based on the size
    of the step. Default: 1e-12

    *itnlimit* is the maximum number of steps to take before stopping.
    Default: 2000

    *abort_test* is a function which tests whether the user has requested
    abort. Default: None.

    *monitor(x,fx,step)* is called every iteration so that a user interface
    function can monitor the progress of the fit.  Default: lambda \*\*kw: True


    Returns the fit result as a dictionary:

    *status* is a status code indicating why the fit terminated.  Turn the
    status code into a string with *STATUS[result.status]*.  Status values
    vary from 1 to 9, with 1 and 2 indicating convergence and the remaining
    codes indicating some form of premature termination.

    *x* is the minimum point

    *fx* is the value fn(x) at the minimum

    *H* is the approximate Hessian matrix, which is the inverse of the
    covariance matrix

    *L* is the cholesky decomposition of H+D, where D is a small correction
    to force H+D to be positive definite.  To compute parameter uncertainty

    *iterations* is the number of iterations

    *evals* is the number of function evaluations

    *linesearch_evals* is the number of function evaluations for line search
    """
    # print("starting QN")
    # If some input parameters are not specified, define default values for them
    # here. First and second parameters fn and x0 must be defined, others may be
    # passed.  If you want to set a value to a parameter, say to typf, make
    # sure all the parameters before this parameter are specified, in this
    # case fn, x0, grad, and Sx if you want to have default values for grad
    # and Sx, for each enter [].
    # important for also computing fcount (function count)
    n = len(x0)
    if x0 is None:
        x0 = zeros(n)

    if grad is None:
        analgrad = 0
    else:
        analgrad = 1

    if Sx is None:
        Sx = ones(n)
        # Sx = x0 + (x0==0.)
    elif len(Sx) != n:
        raise ValueError("sizes of x0 and Sx must be the same")

    if macheps is None:
        # PAK: use finfo rather than macheps
        macheps = finfo("d").eps

    if eta is None:
        eta = macheps

    fcount = 0  # total function count
    fcount_ls = 0  # funciton count due to line search

    # If analytic gradient is available then fn will return both function
    # value and analytic gradient.  Otherwise, use finite difference method
    # for estimating the gradient
    if analgrad == 1:
        fc = fn(x0)
        gc = grad(x0)
        fcount = fcount + 1
    else:
        fc = fn(x0)
        gc = fdgrad(n, x0, fc, fn, Sx, eta)
        fcount = fcount + n + 1

    # Check if the initial guess is a local minimizer
    termcode = umstop0(n, x0, fc, gc, Sx, typf, gradtol)
    consecmax = 0

    # Value to return if we fail early
    # Approximately x0 is a critical point
    xf = x0
    ff = fc
    H = L = None
    if termcode == 0:
        H = inithessunfac(n, fc, typf, Sx)

    # STEP 9.
    xc = x0

    # Iterate until convergence in the following loop
    itncount = 0
    while termcode == 0:  # todo. increase itncount
        # print("update",itncount)
        itncount = itncount + 1

        # disp(['Iteration = ' num2str(itncount)])
        # Find Newton step sN
        H, L = modelhess(n, Sx, macheps, H)
        # the vector obtained in the middle
        middle_step_v = linalg.solve(L, -gc)
        sN = linalg.solve(L.transpose(), middle_step_v)  # the last step
        if isnan(sN).any():
            # print("H",H)
            # print("L",L)
            # print("v",middle_step_v)
            # print("Sx",Sx)
            # print("gc",gc)
            termcode = 9
            break

        # Perform line search (Alg.6.3.1). todo. put param order as in the book
        # print("calling linesearch",xc,fc,gc,sN,Sx,H,L,middle_step_v)
        # print("linesearch",xc,fc)
        retcode, xp, fp, maxtaken, fcnt = linesearch(fn, n, xc, fc, gc, sN, Sx, maxstep, steptol)
        fcount += fcnt
        fcount_ls += fcnt
        # plot(xp(1), xp(2), 'g.')

        # Evaluate gradient at new point xp
        if analgrad == 1:
            gp = grad(xp)
        else:
            gp = fdgrad(n, xp, fp, fn, Sx, eta)
            fcount = fcount + n

        # Check stopping criteria (alg.7.2.1)
        consecmax = consecmax + 1 if maxtaken else 0
        termcode = umstop(n, xc, xp, fp, gp, Sx, typf, retcode, gradtol, steptol, itncount, itnlimit, consecmax)

        if abort_test():
            termcode = 6

        # STEP 10.6
        # If termcode is larger than zero, we found a point satisfying one
        # of the termination criteria, return from here.  Otherwise evaluate
        # the next Hessian approximation (Alg. 9.4.1).
        if termcode > 0:
            xf = xp  # x final
            ff = fp  # f final

        elif not monitor(x=xp, fx=fp, step=itncount):
            termcode = 6

        else:
            H = bfgsunfac(n, xc, xp, gc, gp, macheps, eta, analgrad, H)
            xc = xp
            fc = fp
            gc = gp
        # STOPHERE

    result = dict(status=termcode, x=xf, fx=ff, H=H, L=L, iterations=itncount, evals=fcount, linesearch_evals=fcount_ls)
    # print("result",result, steptol, macheps)
    return result


# ------------------------------------------------------------------------------
# @author: Ismet Sahin
# Alg. 9.4.1

# NOTE:
# BFCG Hessian update is performed unless the following two conditions hold
#    (i) y'*s < sqrt(macheps)*norm(s)*norm(y)
#    (ii)


def bfgsunfac(n, xc, xp, gc, gp, macheps, eta, analgrad, H):
    s = xp - xc
    y = gp - gc
    temp1 = inner(y, s)
    # ISMET : I added condition of having temp1 != 0
    if temp1 >= sqrt(macheps) * linalg.norm(s) * linalg.norm(y) and temp1 != 0:
        if analgrad == 1:
            tol = eta
        else:
            tol = sqrt(eta)

        # deal with noise levels in y
        skipupdate = 1
        t = dot(H, s)
        temp_logicals = abs(y - t) >= tol * maximum(abs(gc), abs(gp))
        if sum(temp_logicals):
            skipupdate = 0

        # do the BFGS update if skipdate is false
        if skipupdate == 0:
            temp2 = dot(s, t)
            H = H + outer(y, y) / temp1 - outer(t, t) / temp2

    return H


# ------------------------------------------------------------------------------
"""
@author: Ismet Sahin
"""


def choldecomp(n, H, maxoffl, macheps):
    minl = (macheps) ** (0.25) * maxoffl

    if maxoffl == 0:
        # H is known to be a positive definite matrix
        maxoffl = sqrt(max(abs(diag(H))))

    minl2 = sqrt(macheps) * maxoffl

    # 3. maxadd is the number (R) specifying the maximum amount added to any
    # diagonal entry of Hessian matrix H
    maxadd = 0

    # 4. form column j of L
    L = zeros((n, n))
    for j in range(1, n + 1):
        L[j - 1, j - 1] = H[j - 1, j - 1] - sum(L[j - 1, 0 : j - 1] ** 2)
        minljj = 0
        for i in range(j + 1, n + 1):
            L[i - 1, j - 1] = H[j - 1, i - 1] - sum(L[i - 1, 0 : j - 1] * L[j - 1, 0 : j - 1])
            minljj = max(abs(L[i - 1, j - 1]), minljj)

        # 4.4
        minljj = max(minljj / maxoffl, minl)

        # 4.5
        if L[j - 1, j - 1] > minljj**2:
            # normal Cholesky iteration
            L[j - 1, j - 1] = sqrt(L[j - 1, j - 1])
        else:
            # augment H[j-1,j-1]
            if minljj < minl2:
                minljj = minl2  # occurs only if maxoffl = 0

            maxadd = max(maxadd, minljj**2 - L[j - 1, j - 1])
            L[j - 1, j - 1] = minljj

        # 4.6
        L[j:n, j - 1] = L[j:n, j - 1] / L[j - 1, j - 1]

    return L, maxadd


# ------------------------------------------------------------------------------
# ALGORITHM 5.6.3

# Ismet Sahin

# function g = fdgrad(n, xc, fc, objfunc, sx, eta)
# g = fdgrad(@obj_function1, 2, [1 -1]', 10, [1 1], eps)

# NOTATION:
#    N : Natural number
#    R : Real number
#    Rn: nx1 real vector
#    Rnxm : nxm real matrix

# INPUTS:
#    n  : the dimension of the gradient vector (N)
#    xc : the current point at which the value of gradient is computed (Rn)
#    fc : function value at xc (R)
#    objfunc : a function handle which is used to compute function values
#    Sx : a n-dim vector, jth entry specifies the typical value of jth param.
# (Rn)
#    eta: equals to 1e-DIGITS where DIGITS is an integer specifying the
# number of reliable digits (R)
# OUTPUT:
#    g : the n-dim finite difference gradient vector (Rn)

# NOTES :
#    hj : is the constant specifying the step size in the direction of jth
# coordinate (R)
#    ej : the unit vector, jth column of the identity matrix (Rn)

# COMMENTS:
# --- FIND STEP SIZE hj
#    1.a : sign(x) does not work for us when x = 0 since this makes the step
# size hj zero which is not allowed. (Step size = 0 => gj = inf.)
#    1.b : evaluation of the step size
#    1.c : a trick to reduce error due to finite precision.  The line xc(j) =
# xc(j) + hj is equivalent to xc = xc + hj * ej where ej is the jth column
# of identity matrix.
#
# --- EVALUATE APPR. GRADIENT
# First evaluate function at xc + hj * ej and then estimate jth entry of
# the gradient.


def fdgrad(n, xc, fc, fn, Sx, eta):
    # create memory for gradient
    g = zeros(n)

    sqrteta = sqrt(eta)
    for j in range(1, n + 1):
        # --- FIND STEP SIZE hj
        if xc[j - 1] >= 0:
            signxcj = 1
        else:
            signxcj = -1  # 1.a

        # 1.b
        hj = sqrteta * max(abs(xc[j - 1]), 1 / Sx[j - 1]) * signxcj

        # 1.c
        tempj = xc[j - 1]
        xc[j - 1] = xc[j - 1] + hj
        hj = xc[j - 1] - tempj

        # --- EVALUATE APPR. GRADIENT
        fj = fn(xc)
        # PAK: hack for infeasible region: point the other way
        if isinf(fj):
            fj = fc + hj
        g[j - 1] = (fj - fc) / hj
        # if isinf(g[j-1]):
        #    print("fc,fj,hj,Sx,xc",fc,fj,hj,Sx[j-1],xc[j-1])

        # now reset the current
        xc[j - 1] = tempj

    # print("gradient", g)
    return g


# ------------------------------------------------------------------------------
# @author: Ismet Sahin
# Example call:
# H = inithessunfac(2, f, 1, [1 0.1]')


def inithessunfac(n, f, typf, Sx):
    temp = max(abs(f), typf)
    H = diag(temp * Sx**2)
    return H


# ------------------------------------------------------------------------------


def linesearch(cost_func, n, xc, fc, g, p, Sx, maxstep, steptol):
    """
    ALGORITHM 6.3.1

    Ismet Sahin

    THE PURPOSE

        is to find a step size which yields the new function value smaller than the
        current function value, i.e. f(xc + alfa*p) <= f(xc) + alfa * lambda * g'p

    CONDITIONS

        g'p < 0
        alfa < 0.5

    NOTATION:
        N : Natural number
        R : Real number
        Rn: nx1 real vector
        Rnxm : nxm real matrix
        Str: a string

    INPUTS
        n : dimensionality (N)
        xc : the current point ( Rn)
        fc : the function value at xc (R)
        obj_func : the function handle to evaluate function values (str like :
           '@costfunction1')
        g : gradient (Rn)
        p : the descent direction (Rn)
        Sx : scale factors (Rn)
        maxstep : maximum step size allowed (R)
        steptol : step tolerance in order to break infinite loop in line search (R)

    OUTPUTS
        retcode : boolean indicating a new point xp found (0) or not (1)    (N).
        xp : the new point (Rn)
        fp : function value at xp (R)
        maxtaken : boolean (N)

    NOTES:
        alfa : is used to prevent function value reductions which are too small.
           Here we'll use a very small number in order to accept very small
           reductions but not too small.
    """

    maxtaken = 0

    # alfa specifies how much function value reduction is allowable.  The
    # smaller the alfa, the smaller the function value reduction we allow.
    alfa = 1e-4

    # the magnitude of the Newton step
    Newtlen = linalg.norm(Sx * p)

    if Newtlen > maxstep:
        # Newton step is larger than the max acceptable step size (maxstep).
        # Make it equal or smaller than maxstep
        p = p * (maxstep / Newtlen)
        Newtlen = maxstep

    initslope = inner(g, p)

    # "Relative length of p as calculated in the stopping routine"
    # rellength = amax(abs(p) / maximum(abs(xc), Sx))    # this was a bug
    rellength = amax(abs(p) / maximum(abs(xc), 1 / Sx))

    minlambda = steptol / rellength

    lambdaM = 1.0

    # In this loop, we try to find an acceptable next point
    # xp = xc + lambda * p by finding an optimal lambda based on one
    # dimensional quadratic and cubic models
    fcount = 0
    while True:  # 10 starts.
        # next point candidate
        xp = xc + lambdaM * p
        if isnan(xp).any():
            # print("nan xp")
            retcode = 1
            xp, fp = xc, fc
            break
        if fcount > 20:
            # print("too many cycles in linesearch",xp)
            retcode = 2
            xp, fp = xc, fc
            break
        # function value at xp
        fp = cost_func(xp)
        # print("linesearch",fcount,xp,xc,lambdaM,p,fp,fc)
        if isinf(fp):
            fp = 2 * fc  # PAK: infeasible region hack
        fcount = fcount + 1
        if fp <= fc + alfa * lambdaM * initslope:
            # satisfactory xp is found
            retcode = 0
            if lambdaM == 1.0 and Newtlen > 0.99 * maxstep:
                maxtaken = 1
            # return from here
            break
        elif lambdaM < minlambda:
            # step length is too small, so a satisfactory xp cannot be found
            # print("step",lambdaM,minlambda,steptol,rellength)
            retcode = 3
            xp, fp = xc, fc
            break
        else:  # 10.3c starts
            # reduce lambda by a factor between 0.1 and 0.5
            if lambdaM == 1.0:
                # first backtrack with one dimensional quadratic fit
                lambda_temp = -initslope / (2.0 * (fp - fc - initslope))
                # print("L1",lambda_temp)
            else:
                # perform second and following backtracks with cubic fit
                Mt = array(
                    [[1.0 / lambdaM**2, -1.0 / lambda_prev**2], [-lambda_prev / lambdaM**2, lambdaM / lambda_prev**2]]
                )
                vt = array([[fp - fc - lambdaM * initslope], [fp_prev - fc - lambda_prev * initslope]])
                ab = (1.0 / (lambdaM - lambda_prev)) * dot(Mt, vt)
                # a = ab(1) and b = ab(2)
                disc = ab[1, 0] ** 2 - 3.0 * ab[0, 0] * initslope
                # print("Mt,vt,ab,disc",Mt,vt,ab,disc)
                if ab[0, 0] == 0.0:
                    # cubic model turn out to be a quadratic
                    lambda_temp = -initslope / (2.0 * ab[1, 0])
                    # print("L2",lambda_temp)
                else:
                    # the model is a legitimate cubic
                    lambda_temp = (-ab[1, 0] + sqrt(disc)) / (3.0 * ab[0, 0])
                    # print("L3",lambda_temp)

                if lambda_temp > 0.5 * lambdaM:
                    # larger than half of previous lambda is not allowed.
                    lambda_temp = 0.5 * lambdaM
                    # print("L4",lambda_temp)

            lambda_prev = lambdaM
            fp_prev = fp
            if lambda_temp <= 0.1 * lambdaM:
                # smaller than 1/10 th of previous lambda is not allowed.
                lambdaM = 0.1 * lambdaM
            else:
                lambdaM = lambda_temp

            # print('lambda = ', lambdaM)

    # return xp, fp, retcode
    return retcode, xp, fp, maxtaken, fcount


# ------------------------------------------------------------------------------


# @author: Ismet Sahin
# ALGORITHM 1.3.1
def machineeps():
    macheps = 1.0
    while (macheps + 1) != 1:
        macheps = macheps / 2

    macheps = 2 * macheps
    return macheps


# ------------------------------------------------------------------------------


def modelhess(n, Sx, macheps, H):
    """
    @author: Ismet Sahin.
    Thanks to Christopher Meeting for his help in converting this module from
    Matlab to Python

    ALGORITHM 5.5.1

    NOTES:
        Currently we are not implementing steps 1, 14, and 15 (TODO)

    This function performs perturbed Cholesky decomposition (CD) as if the input
    Hessian matrix is positive definite.  The code for perturbed CD resides in
    choldecomp.m file which returns the factored lower triangle matrix L and a
    number, maxadd, specifying the largest number added to a diagonal element of
    H during the CD decomposition.  This function checks if the decomposition is
    completed without adding any positive number to the diagonal elements of H,
    i.e. maxadd <= 0.  Otherwise, this function adds the least number to the
    diagonals of H which makes it positive definite based on maxadd and other
    entries in H.
    EXAMPLE CALLS::

             A1 =[2     0    2.4
                  0     2     0
                  2.4     0     3]

             A2 =[2     0    2.5
                   0     2     0
                  2.5     0     3]

             A3 =[2     0    10
                   0     2     0
                  10     0     3]
    """

    # SCALING
    scale_needed = 0  # ISMET uses this parameter
    if sum(Sx - ones(n)) != 0:
        # scaling is requested by the user
        scale_needed = 1
        Dx = diag(Sx)
        Dx_inv = diag(1.0 / Sx)
        H = dot(Dx_inv, dot(H, Dx_inv))

    # STEP I.
    sqrteps = sqrt(macheps)

    # 2-4.
    H_diag = diag(H)
    maxdiag = max(H_diag)
    mindiag = min(H_diag)

    # 5.
    maxposdiag = max(0, maxdiag)

    # 6. mu is the amount to be added to diagonal of H before the
    # Cholesky decomp. If the minimum diagonal is much much smaller than
    # the maximum diagonal element then adjust mu accordingly otherwise mu = 0.
    if mindiag <= sqrteps * maxposdiag:
        mu = 2 * (maxposdiag - mindiag) * sqrteps - mindiag
        maxdiag = maxdiag + mu
    else:
        mu = 0

    # 7. maximum of off-diagonal elements of H
    diag_infinite = diag(inf * ones(n))
    maxoff = (H - diag_infinite).max()

    # 8. if maximum off diagonal element is much much larger than the maximum
    # diagonal element of the Hessian H
    if maxoff * (1 + 2 * sqrteps) > maxdiag:
        mu = mu + (maxoff - maxdiag) + 2 * sqrteps * maxoff
        maxdiag = maxoff * (1 + 2 * sqrteps)

    # 9.
    if maxdiag == 0:  # if H == 0
        mu = 1
        maxdiag = 1

    # 10. mu>0 => need to add mu amount to the diagonal elements: H = H + mu*I
    if mu > 0:
        diag_mu = diag(mu * ones(n))
        H = H + diag_mu

    # 11.
    maxoffl = sqrt(max(maxdiag, maxoff / n))

    # STEP II. Perform perturbed Cholesky decomposition H + D = LL' where D is
    # a diagonal matrix which is implicitly added to H if H is not positive
    # definite. Matrix D has only positive elements. The output variable maxadd
    # indicates the maximum number added to a diagonal entry of the Hesian,
    # i.e. the maximum of D. If maxadd is returned 0, then H was indeed pd
    # and L is the resulting factor.
    # 12.
    L, maxadd = choldecomp(n, H, maxoffl, macheps)

    # STEP III.
    # 13. If maxadd <= 0, we are done H was positive definite.
    if maxadd > 0:
        # H was not positive definite
        # print('WARNING: Hessian is not pd. Max number added to H is ',maxadd)
        maxev = H[0, 0]
        minev = H[0, 0]
        for i in range(1, n + 1):
            offrow = sum(abs(H[0 : i - 1, i - 1])) + sum(abs(H[i - 1, i:n]))
            maxev = max(maxev, H[i - 1, i - 1] + offrow)
            minev = min(minev, H[i - 1, i - 1] - offrow)

        sdd = (maxev - minev) * sqrteps - minev
        sdd = max(sdd, 0)
        mu = min(maxadd, sdd)
        H = H + diag(mu * ones(n))
        L, maxadd = choldecomp(n, H, 0, macheps)

    if scale_needed:  # todo. this calculation can be done faster
        H = dot(Dx, dot(H, Dx))
        L = dot(Dx, L)

    return H, L


# ------------------------------------------------------------------------------
def umstop(n, xc, xp, f, g, Sx, typf, retcode, gradtol, steptol, itncount, itnlimit, consecmax):
    """
    #@author: Ismet Sahin

    ALGORITHM 7.2.1

    Return codes:
    Note that return codes are nonnegative integers. When it is not zero, there is
    a termination condition which is satisfied.
       0 : None of the termination conditions is satisfied
       1 : Magnitude of scaled grad is less than gradtol; this is the primary
           condition. The new point xp is most likely a local minimizer.  If gradtol
           is too large, then this condition can be satisfied easier and therefore
           xp may not be a local minimizer
       2 : Scaled distance between last two points is less than steptol; xp might be
           a local minimizer.  This condition may also be satisfied if step is
           chosen too large or the algorithm is far from the minimizer and making
           small progress
       3 : The algorithm cannot find a new point giving smaller function value than
           the current point.  The current may be a local minimizer, or analytic
           gradient implementation has some mistakes, or finite difference gradient
           estimation is not accurate, or steptol is too large.
       4 : Maximum number of iterations are completed
       5 : The maximum step length maxstep is taken for last ten consecutive
           iterations.  This may happen if the function is not bounded from below,
           or the function has a finite asymptote in some direction, or maxstep is
           too small.
    """

    termcode = 0
    if retcode == 1:
        termcode = 3
    elif retcode == 2:
        termcode = 7
    elif retcode == 3:
        termcode = 8
    elif retcode > 0:
        raise ValueError("Unknown linesearch return code")
    elif max(abs(g) * maximum(abs(xp), 1 / Sx) / max(abs(f), typf)) <= gradtol:
        # maximum component of scaled gradient is smaller than gradtol.
        # TODO: make sure not to use a too large typf value which leads to the
        # satisfaction of this algorithm easily.
        termcode = 1
    elif max(abs(xp - xc) / maximum(abs(xp), 1 / Sx)) <= steptol:
        # maximum component of scaled step is smaller than steptol
        termcode = 2
    elif itncount >= itnlimit:
        # maximum number of iterations are performed
        termcode = 4
    elif consecmax == 10:
        # not more than 10 steps will be taken consecutively.
        termcode = 5

    return termcode


# ------------------------------------------------------------------------------
# @author: Ismet Sahin

# This function checks whether initial conditions are acceptable for
# continuing unconstrained optimization

# f : the function value at x0, i.e. f = f(x0),  (R)
# g : the gradient at x0, (Rn)

# termcode = 0 : x0 is not a critical point of f(x), (Z)
# termcode = 1 : x0 is a critical point of f(x), (Z)

# Note that x0 may be a critical point of the function; in this case, it is
# either a local minimizer or a saddle point of the function.  If the Hessian
# at x0 is positive definite than it is indeed a local minimizer.  Instead of
# checking Hessian, we can also restart the driver program umexample from
# another point which is close to x0.  If x0 is the local minimizer, the
# algorithm will approach it.


def umstop0(n, x0, f, g, Sx, typf, gradtol):
    # consecmax = 0
    if max(abs(g) * maximum(abs(x0), 1.0 / Sx) / max(abs(f), typf)) <= 1e-3 * gradtol:
        termcode = 1
    else:
        termcode = 0
    return termcode


# ------------------------------------------------------------------------------


def example_call():
    print("***********************************")

    # Rosenbrock function
    fn = lambda p: (1 - p[0]) ** 2 + 100 * (p[1] - p[0] ** 2) ** 2
    grad = lambda p: array([-2 * (1 - p[0]) - 400 * (p[1] - p[0] ** 2) * p[0], 200 * (p[1] - p[0] ** 2)])
    x0 = array([2.320894, -0.534223])
    # x0 = array([2.0,1.0])

    result = quasinewton(fn=fn, x0=x0, grad=grad)
    # result = quasinewton(fn=fn, x0=x0)

    print("\n\nInitial point x0 = ", x0, ", f(x0) = ", fn(x0))
    for k in sorted(result.keys()):
        print(k, "=", result[k])


if __name__ == "__main__":
    example_call()



"""
Build a bumps model from a function and data.

Example
-------

Given a function *sin_model* which computes a sine wave at times *t*::

    from numpy import sin
    def sin_model(t, freq, phase):
        return sin(2*pi*(freq*t + phase))

and given data *(y,dy)* measured at times *t*, we can define the fit
problem as follows::

    from bumps.names import *
    M = Curve(sin_model, t, y, dy, freq=20)

The *freq* and *phase* keywords are optional initial values for the model
parameters which otherwise default to zero.  The model parameters can be
accessed as attributes on the model to set fit range::

    M.freq.range(2, 100)
    M.phase.range(0, 1)

As usual, you can initialize or assign parameter expressions to the the
parameters if you want to tie parameters together within or between models.

Note: there is sometimes difficulty getting bumps to recognize the function
during fits, which can be addressed by putting the definition in a separate
file on the python path.  With the windows binary distribution of bumps,
this can be done in the problem definition file with the following code::

    import os
    from bumps.names import *
    sys.path.insert(0, os.getcwd())

The model function can then be imported from the external module as usual::

    from sin_model import sin_model
"""

__all__ = ["Curve", "PoissonCurve", "plot_err"]

import inspect
import warnings
from typing import Callable, Literal
from copy import deepcopy

import numpy as np
from numpy import log, pi, sqrt

from .parameter import Parameter


def _parse_pars(fn, init=None, skip=0, name=""):
    """
    Extract parameter names from function definition.

    *fn* is the function definition.  This could be declared as
    *fn(p1, p2, p3, ...)* where *p1*, etc. are the fittable parameters.

    *init* is a dictionary of initial values for the parameters,
    overriding any default values.  If called from a constructor with
    **kwargs representing unknown named arguments, use *init=kwargs*.

    *skip* is the number of parameters to skip.  This will be *skip=0*
    for a function which defines the log likelihood directly or one
    that returns a set of residuals. For parameterized curves such as
    *fn(x, p1, p2, ...)* use *skip=1*.  For surfaces with
    *fn(x, y, p1, p2, ...)* use *skip=2*.

    *name* is added to each parameter name to differentiate it from other
    parameters in the same fit.

    A default value in the function definition such as *pk=value* will
    be set as the default value for the parameter.  If the default is
    *pk=None* then the parameter will be non-fittable, and instead set
    through *init*.
    """
    sig = inspect.signature(fn)
    params = sig.parameters.values()
    pnames = [p.name for p in params]

    valid = [p.kind in (inspect.Parameter.POSITIONAL_ONLY, inspect.Parameter.POSITIONAL_OR_KEYWORD) for p in params]
    if not all(valid):
        raise TypeError(f"Only positional and keyword arguments allowed for {fn.__name__}")

    # TODO: need "self" handling for passed methods
    # Skip the first argument if it is x or maybe skip x, y.
    pnames = pnames[skip:]

    # Parameters default to zero
    defaults = dict((p, 0) for p in pnames)

    # If the function provides default values, use those.
    for param in list(params)[skip:]:
        if param.default is not inspect.Parameter.empty:
            defaults[param.name] = param.default

    # Non-fittable parameters need to be sent in as None
    state_vars = set(p for p, v in defaults.items() if v is None)

    # Regardless, use any values specified in the constructor, but first
    # check that they exist as function parameters.
    invalid = set(init.keys()) - set(pnames)
    if invalid:
        raise TypeError("Invalid initializers: %s" % ", ".join(sorted(invalid)))
    defaults.update(init)

    # Build parameters out of ranges and initial values
    # maybe:  name=(p+name if name.startswith('_') else name+p)
    pars = dict((p, Parameter.default(defaults[p], name=name + p)) for p in pnames if p not in state_vars)

    state = dict((p, v) for p, v in defaults.items() if p in state_vars)

    # print("pars", pars)
    # print("state", state)
    return pars, state


def _assign_pars(obj, pars):
    # Make parameters accessible as model attributes
    for k, v in pars.items():
        if hasattr(obj, k):
            raise TypeError("Parameter cannot be named %s" % k)
        setattr(obj, k, v)


class Curve(object):
    r"""
    Model a measurement with a user defined function.

    The function *fn(x,p1,p2,...)* should return the expected value *y* for
    each point *x* given the parameters *p1*, *p2*, etc.  *dy* is the
    uncertainty for each measured value *y*.  If not specified, it defaults
    to 1. Multi-valued functions, which return multiple *y* values for each
    *x* value, should have *x* as a vector of length *n* and *y*, *dy* as
    arrays of size *[n, k]*.

    Initial values for the parameters can be set as *p=value* arguments to
    *Curve*. If no value is set, then the initial value will be taken from
    the default value given in the definition of *fn*, or set to 0 if the
    parameter is not defined with an initial value.  Arbitrary non-fittable
    data can be passed to the function as parameters, but only if the
    parameter is given a default value of *None* in the function definition,
    and has the initial value set as an argument to *Curve*.  Defining
    *state=dict(key=value, ...)* before *Curve*, and calling *Curve* as
    *Curve(..., \*\*state)* works pretty well.

    *Curve* takes the following special keyword arguments:

    * *name* is added to each parameter name when the parameter is defined.
      The filename for the data is a good choice, since this allows you to keep
      the parameters straight when fitting multiple datasets simultaneously.

    * *plot* is an alternative plotting function. The function should be
      defined as *plot(x,y,dy,fy,\*\*kw)*. The keyword arguments will be
      filled with the values of the parameters used to compute *fy*.  It
      will be easiest to list the parameters you need to make your plot
      as positional arguments after *x,y,dy,fy* in the plot function
      declaration.  For example, *plot(x,y,dy,fy,p3,\*\*kw)* will make the
      value of parameter *p3* available as a variable in your function.  The
      special keyword *view* will be a string containing *linear*, *log*,
      *logx*, or *loglog*.  If only showing the residuals, the string
      will be *residual*.

    * *plot_x* is an array giving the sample points to use when plotting
      the theory function, if different from the *x* values at which the
      function is sampled.  Use this to draw a smooth curve between the
      fitted points.  This value is ignored if you provide your own plot
      function.

    * *labels* are the axis labels for the plot.  This should include
      units in parentheses. If the function is multi-valued then
      use *['x axis', 'y axis', 'line 1', 'line 2', ...]*.

    The data uncertainty is assumed to follow a gaussian distribution.
    If measurements draw from some other uncertainty distribution, then
    subclass Curve and replace nllf with the correct probability given the
    residuals.  See the implementation of :class:`PoissonCurve` for an example.
    """

    def __init__(self, fn, x, y, dy=None, name="", labels=None, plot=None, plot_x=None, **kwargs):
        self.x, self.y = np.asarray(x), np.asarray(y)
        if dy is None:
            self.dy = 1
        else:
            self.dy = np.asarray(dy)
            if (self.dy <= 0).any():
                raise ValueError("measurement uncertainty must be positive")

        if len(self.x.shape) == 1 and len(self.y.shape) > 1:
            num_curves = self.y.shape[0]
        else:
            num_curves = 1
        self._num_curves = num_curves  # use same value everywhere

        # interpret labels parameter
        if labels is None:
            labels = ["x", "y"]
        elif len(labels) < 2 or len(labels) != num_curves + 2:
            if num_curves > 1:
                lines = "line1, ..., line%d" % num_curves
            else:
                lines = "line"
            raise TypeError("labels should be [x, y, %s]" % lines)

        if len(labels) == 2:
            if num_curves > 1:
                line_labels = ["y%d" % k for k in range(num_curves)]
            else:
                line_labels = [labels[1]]
            labels = list(labels) + line_labels
        self.labels = labels

        # TODO: self.fn is a duplicate of self._function below. Deprecated?
        self.fn = fn
        self.name = name  # if name else fn.__name__ + " "
        self.plot_x = plot_x

        pars, state = _parse_pars(fn, init=kwargs, skip=1, name=name)

        # Make parameters accessible as model attributes
        _assign_pars(self, pars)
        # _assign_pars(state, self)  # ... and state variables as well

        # Remember the function, parameters, and number of parameters
        # Note: we are remembering the parameter names and not the
        # parameters themselves so that the caller can tie parameters
        # together using model1.par = model2.par.  Otherwise we would
        # need to override __setattr__ to intercept assignment to the
        # parameter attributes and redirect them to the a _pars dictionary.
        # ... and similarly for state if we decide to make them attributes.
        self._function = fn
        self._pnames = list(sorted(pars.keys()))
        self._state = state
        self._plot = plot
        self._cached_theory = None
        self._webview_plots = {}

    def update(self):
        self._cached_theory = None

    def parameters(self):
        return dict((p, getattr(self, p)) for p in self._pnames)

    def numpoints(self):
        return np.prod(self.y.shape)

    def theory(self, x=None):
        # Use cache if x is None, otherwise compute theory with x.
        if x is None:
            if self._cached_theory is None:
                self._cached_theory = self._compute_theory(self.x)
            return self._cached_theory
        return self._compute_theory(x)

    def _compute_theory(self, x):
        kw = self._fetch_pars()
        return self._function(x, **kw)

    def _fetch_pars(self):
        kw = dict((p, getattr(self, p).value) for p in self._pnames)
        kw.update(self._state)
        return kw

    def simulate_data(self, noise=None):
        theory = self.theory()
        if noise is not None:
            if noise == "data":
                pass
            elif noise < 0:
                self.dy = -0.01 * noise * theory
            else:
                self.dy = noise
        self.y = theory + np.random.randn(*theory.shape) * self.dy

    def residuals(self):
        return (self.theory() - self.y) / self.dy

    def nllf(self):
        r = self.residuals()
        return 0.5 * np.sum(r**2)

    def save(self, basename):
        # TODO: need header line with state vars as json
        # TODO: need to support nD x,y,dy
        if len(self.x.shape) > 1:
            warnings.warn("Save not supported for nD x values")
            return

        theory = self.theory()
        if self._num_curves > 1:
            # Multivalued y, dy for single valued x.
            columns = [self.x]
            headers = ["x"]
            for k, (y, dy, fx) in enumerate(zip(self.y, self.dy, theory)):
                columns.extend((y, dy, fx))
                headers.extend(("y[%d]" % (k + 1), "dy[%d]" % (k + 1), "fx[%d]" % (k + 1)))
        else:
            # Single-valued y, dy for single valued x.
            headers = ["x", "y", "dy", "fy"]
            columns = [self.x, self.y, self.dy, theory]
        data = np.vstack(columns)
        outfile = basename + ".dat"
        with open(outfile, "w") as fd:
            fd.write("# " + "\t ".join(headers) + "\n")
            np.savetxt(fd, data.T)

    def plot(self, view=None):
        if self._plot is not None:
            kw = self._fetch_pars()
            self._plot(self.x, self.y, self.dy, self.theory(), view=view, **kw)
            return

        import pylab
        from .plotutil import coordinated_colors

        x = self.x
        if self.plot_x is not None:
            theory_x, theory_y = self.plot_x, self.theory(self.plot_x)
        else:
            theory_x, theory_y = x, self.theory()
        resid = self.residuals()

        if self._num_curves > 1:
            y, dy, theory_y, resid = self.y.T, self.dy.T, theory_y.T, resid.T
        else:
            y, dy, theory_y, resid = (v[:, None] for v in (self.y, self.dy, theory_y, resid))

        colors = tuple(coordinated_colors() for _ in range(self._num_curves))
        labels = self.labels

        # print "kw_plot",kw
        if view == "residual":
            _plot_resids(x, resid, colors, labels=labels, view=view)
        else:
            plot_ratio = 4
            h = pylab.subplot2grid((plot_ratio, 1), (0, 0), rowspan=plot_ratio - 1)
            for tick_label in h.get_xticklabels():
                tick_label.set_visible(False)
            _plot_fits(data=(x, y, dy), theory=(theory_x, theory_y), colors=colors, labels=labels, view=view)
            # pylab.gca().xaxis.set_visible(False)
            # pylab.gca().spines['bottom'].set_visible(False)
            # pylab.gca().set_xticks([])

            pylab.subplot2grid((plot_ratio, 1), (plot_ratio - 1, 0), sharex=h)
            _plot_resids(x, resid, colors=colors, labels=labels, view=view)

    def register_webview_plot(
        self, plot_title: str, plot_function: Callable, change_with: Literal["parameter", "uncertainty"]
    ):
        # Plot function syntax: f(model, problem, state)
        # change_with = 'parameter' or 'uncertainty'

        self._webview_plots[plot_title] = dict(func=plot_function, change_with=change_with)

    @property
    def webview_plots(self):
        return self._webview_plots


def _plot_resids(x, resid, colors, labels, view):
    import pylab

    pylab.axhline(y=1, ls="dotted", color="k")
    pylab.axhline(y=0, ls="solid", color="k")
    pylab.axhline(y=-1, ls="dotted", color="k")
    for k, color in enumerate(colors):
        pylab.plot(x, resid[:, k], ".", color=color["base"])
    pylab.gca().locator_params(axis="y", tight=True, nbins=4)
    pylab.xlabel(labels[0])
    pylab.ylabel("(f(x)-y)/dy")
    if view == "logx":
        pylab.xscale("log")
    elif view == "loglog":
        pylab.xscale("log")


def _plot_fits(data, theory, colors, labels, view):
    import pylab

    x, y, dy = data
    theory_x, theory_y = theory
    for k, color in enumerate(colors):
        pylab.errorbar(x, y[:, k], yerr=dy[:, k], fmt=".", color=color["base"], label="_")
        pylab.plot(theory_x, theory_y[:, k], "-", color=color["dark"], label=labels[k + 2])
    # Note: no xlabel since it is supplied by the residual plot below this plot
    pylab.ylabel(labels[1])
    if len(colors) > 1:
        pylab.legend()
    if view == "log":
        pylab.xscale("linear")
        pylab.yscale("log")
    elif view == "logx":
        pylab.xscale("log")
        pylab.yscale("linear")
    elif view == "logy":
        pylab.xscale("linear")
        pylab.yscale("log")
    elif view == "loglog":
        pylab.xscale("log")
        pylab.yscale("log")
    else:  # view == 'linear'
        pylab.xscale("linear")
        pylab.yscale("linear")


def plot_resid(x, resid):
    """
    **DEPRECATED**
    """
    import pylab

    pylab.axhline(y=1, ls="dotted", color="k")
    pylab.axhline(y=0, ls="solid", color="k")
    pylab.axhline(y=-1, ls="dotted", color="k")
    pylab.plot(x, resid, ".")
    pylab.gca().locator_params(axis="y", tight=True, nbins=4)
    pylab.ylabel("Residuals")


def plot_err(x, y, dy, fy, view=None, **kw):
    """
    **DEPRECATED**: subclass Curve and override the plot function.

    Plot data *y* and error *dy* against *x*.

    *view* is one of linear, log, logx or loglog.
    """
    import pylab

    pylab.errorbar(x, y, yerr=dy, fmt=".")
    pylab.plot(x, fy, "-")
    if view == "log":
        pylab.xscale("linear")
        pylab.yscale("log")
    elif view == "logx":
        pylab.xscale("log")
        pylab.yscale("linear")
    elif view == "loglog":
        pylab.xscale("log")
        pylab.yscale("log")
    else:  # view == 'linear'
        pylab.xscale("linear")
        pylab.yscale("linear")


_LOGFACTORIAL = np.array([log(np.prod(np.arange(1.0, k + 1))) for k in range(21)])


def logfactorial(n):
    """Compute the log factorial for each element of an array"""
    result = np.empty(n.shape, dtype="double")
    idx = n <= 20
    result[idx] = _LOGFACTORIAL[np.asarray(n[idx], "int32")]
    n = n[~idx]
    result[~idx] = n * log(n) - n + log(n * (1 + 4 * n * (1 + 2 * n))) / 6 + log(pi) / 2
    return result


class PoissonCurve(Curve):
    r"""
    Model a measurement with Poisson uncertainty.

    The nllf is calculated using Poisson probabilities, but the curve itself
    is displayed using the approximation that $\sigma_y \approx \sqrt(y)$.

    See :class:`Curve` for details.
    """

    def __init__(self, fn, x, y, name="", **fnkw):
        dy = sqrt(y) + (y == 0) if y is not None else None
        Curve.__init__(self, fn, x, y, dy, name=name, **fnkw)
        self._logfacty = logfactorial(y) if y is not None else None
        self._logfactysum = np.sum(self._logfacty)

    ## Assume gaussian residuals for now
    # def residuals(self):
    #    # TODO: provide individual probabilities as residuals
    #    # or perhaps the square roots --- whatever gives a better feel for
    #    # which points are driving the fit
    #    theory = self.theory()
    #    return np.sqrt(self.y * log(theory) - theory - self._logfacty)

    def nllf(self):
        theory = self.theory()
        if (theory <= 0).any():
            return 1e308
        return -sum(self.y * log(theory) - theory) + self._logfactysum

    def simulate_data(self, noise=None):
        theory = self.theory()
        self.y = np.random.poisson(theory)
        self.dy = sqrt(self.y) + (self.y == 0)
        self._logfacty = logfactorial(self.y)
        self._logfactysum = np.sum(self._logfacty)



"""

Notes on random numbers
=======================

Uses dream.util.rng as the random number generator.

You can set the seed using::

    dream.util.rng = numpy.random.RandomState(seed)

This interface doesn't feel right, since one instance of DREAM may
influence another if they are running within one another.  Putting
the rng on the dream class may be a better option.
"""

from .model import MCMCModel
from .core import Dream
# from .core import dream
# from .initpop import *  # cov_init, lhs_init
# from .model import *    #
# from .state import *    # load_state, save_state
# from .views import *    # plotting routines
# from .util import console
# from .stats import *



# This program is public domain
# Author Paul Kienzle
"""
2-D correlation histograms

Generate 2-D correlation histograms and display them in a figure.

Uses false color plots of density.
"""

__all__ = ["Corr2d"]

import numpy as np
from numpy import inf

from matplotlib import cm, colors, image, artist
from matplotlib.font_manager import FontProperties
from matplotlib.ticker import MaxNLocator

try:
    COLORMAP = colors.LinearSegmentedColormap.from_list("density", ("w", "y", "g", "b", "r"))
except Exception:
    COLORMAP = cm.gist_earth_r


class Corr2d(object):
    """
    Generate and manage 2D correlation histograms.
    """

    def __init__(self, data, labels=None, **kw):
        if labels is None:
            labels = ["P" + str(i + 1) for i, _ in enumerate(data)]
        self.N = len(data)
        self.labels = labels
        self.data = data
        self.hists = _hists(data, **kw)
        # for k, v in self.hists.items():
        #    print k, (v[1][0], v[1][-1]), (v[2][0], v[2][-1])
        self.ax = None  # will be set on plot

    def R(self):
        return np.corrcoef(self.data)

    def __getitem__(self, key):
        """
        Retrieve correlation histogram for data[i] X data[j].

        Returns bin i edges, bin j edges, and histogram
        """
        i, j = key
        return self.hists[i, j]

    def plot(self, title=None, fig=None):
        """
        Plot the correlation histograms on the specified figure
        """
        import pylab

        if fig is None:
            pylab.clf()
            fig = pylab.gcf()
        if title is not None:
            fig.text(0.5, 0.95, title, horizontalalignment="center", fontproperties=FontProperties(size=16))
        self.ax = _plot(fig, self.hists, self.labels, self.N)


def _hists(data, ranges=None, **kw):
    """
    Generate pair-wise correlation histograms
    """
    n = len(data)
    if ranges is None:
        low, high = np.min(data, axis=1), np.max(data, axis=1)
        ranges = [(l, h) for l, h in zip(low, high)]
    return dict(
        ((i, j), np.histogram2d(data[i], data[j], range=[ranges[i], ranges[j]], **kw))
        for i in range(0, n)
        for j in range(i + 1, n)
    )


def _plot(fig, hists, labels, n, show_ticks=None):
    """
    Plot pair-wise correlation histograms
    """
    if n <= 1:
        fig.text(0.5, 0.5, "No correlation plots when only one variable", ha="center", va="center")
        return
    vmin, vmax = inf, -inf
    for data, _, _ in hists.values():
        positive = data[data > 0]
        if len(positive) > 0:
            vmin = min(vmin, np.amin(positive))
            vmax = max(vmax, np.amax(positive))
    norm = colors.LogNorm(vmin=vmin, vmax=vmax, clip=False)
    # norm = colors.Normalize(vmin=vmin, vmax=vmax)
    mapper = image.FigureImage(fig)
    mapper.set_array(np.zeros((1, 1)))
    mapper.set_cmap(cmap=COLORMAP)
    mapper.set_norm(norm)

    if show_ticks is None:
        show_ticks = n < 3
    ax = {}
    Nr = Nc = n - 1
    for i in range(0, n - 1):
        for j in range(i + 1, n):
            sharex = ax.get((0, j), None)
            sharey = ax.get((i, i + 1), None)
            a = fig.add_subplot(Nr, Nc, (Nr - i - 1) * Nc + j, sharex=sharex, sharey=sharey)
            ax[(i, j)] = a
            a.xaxis.set_major_locator(MaxNLocator(4, steps=[1, 2, 4, 5, 10]))
            a.yaxis.set_major_locator(MaxNLocator(4, steps=[1, 2, 4, 5, 10]))
            data, x, y = hists[(i, j)]
            data = np.clip(data, vmin, vmax)
            a.pcolorfast(y, x, data, cmap=COLORMAP, norm=norm)
            # Show labels or hide ticks
            if i != 0:
                artist.setp(a.get_xticklabels(), visible=False)
            if i == n - 2 and j == n - 1:
                a.set_xlabel(labels[j])
                # a.xaxis.set_label_position("top")
                # a.xaxis.set_offset_position("top")
            if not show_ticks:
                a.xaxis.set_ticks([])
            if j == i + 1:
                a.set_ylabel(labels[i])
            else:
                artist.setp(a.get_yticklabels(), visible=False)
            if not show_ticks:
                a.yaxis.set_ticks([])

            a.zoomable = True

    # Adjust subplots and add the colorbar
    fig.subplots_adjust(left=0.07, bottom=0.07, top=0.9, right=0.85, wspace=0.0, hspace=0.0)
    cax = fig.add_axes([0.88, 0.2, 0.04, 0.6])
    fig.colorbar(mapper, cax=cax, orientation="vertical")
    return ax


def zoom(event, step):
    ax = event.inaxes
    if not hasattr(ax, "zoomable"):
        return

    # TODO: test logscale
    step *= 3

    if ax.zoomable is not True and "mapper" in ax.zoomable:
        mapper = ax.zoomable["mapper"]
        if event.ydata is not None:
            lo, hi = mapper.get_clim()
            pt = event.ydata * (hi - lo) + lo
            lo, hi = _rescale(lo, hi, pt, step)
            mapper.set_clim((lo, hi))
    if ax.zoomable is True and event.xdata is not None:
        lo, hi = ax.get_xlim()
        lo, hi = _rescale(lo, hi, event.xdata, step)
        ax.set_xlim((lo, hi))
    if ax.zoomable is True and event.ydata is not None:
        lo, hi = ax.get_ylim()
        lo, hi = _rescale(lo, hi, event.ydata, step)
        ax.set_ylim((lo, hi))
    ax.figure.canvas.draw_idle()


def _rescale(lo, hi, pt, step):
    scale = float(hi - lo) * step / (100 if step > 0 else 100 - step)
    bal = float(pt - lo) / (hi - lo)
    new_lo = lo - bal * scale
    new_hi = hi + (1 - bal) * scale
    return new_lo, new_hi



# This program is public domain
# Author: Paul Kienzle
"""
Parallel tempering for continuous function optimization and uncertainty analysis.

The program performs Markov chain Monte Carlo exploration of a probability
density function using a combination of random and differential evolution
updates.
"""

__all__ = ["parallel_tempering"]

import numpy as np
from numpy import asarray, zeros, ones, exp, diff, std, inf, array, nonzero, sqrt, zeros_like
from numpy.linalg import norm
from numpy.random import rand, randn, randint, permutation


def every_ten(step, x, fx, P, E):
    if step % 10:
        print(step, fx[step], x[step])


def parallel_tempering(nllf, p, bounds, T=None, steps=1000, CR=0.9, burn=1000, monitor=every_ten, logfile=None):
    r"""
    Perform a MCMC walk using multiple temperatures in parallel.

    :Parameters:

    *nllf* : function(vector) -> float
        Negative log likelihood function to be minimized.  $\chi^2/2$ is a
        good choice for curve fitting with no prior restraints on the possible
        input parameters.
    *p* : vector
        Initial value
    *bounds* : vector, vector
        Box constraints on the parameter values.  No support for indefinite
        or semi-definite programming at present
    *T* : vector | 0 < T[0] < T[1] < ...
        Temperature vector.  Something like logspace(-1,1,10) will give
        you 10 logarithmically spaced temperatures between 0.1 and 10.  The
        maximum temperature T[-1] determines the size of the barriers that
        can be easily jumped.  Note that the number of temperature values
        limits the amount of parallelism available in the algorithm, so it
        may gather statistics more quickly, though it will not necessarily
        converge any faster.
    *steps* = 1000 : int
        Length of the accumulation vector.  The returned history will store
        this many values for each temperature.  These values can be used in
        a weighted histogram to determine parameter uncertainty.
    *burn* = 1000 : int | [0,inf)
        Number of iterations to perform in addition to steps.  Only the
        last *steps* points will be preserved for each temperature.  Since
        the
        value should be in the same order as *steps* to be sure that the
        full history is acquired.
    *CR* = 0.9 : float | [0,1]
        Cross-over ratio.  This is the differential evolution crossover
        ratio to use when computing step size and direction.  Use a small
        value to step through the dimensions one at a time, or a large value
        to step through all at once.
    *monitor* = every_ten : function(int step, vector x, float fx) -> None
        Function to called at every iteration with the step number the
        best point and the best value.
    *logfile* = None : string
        Name of the file which will log the history of every accepted step.
        Note that this includes all of the burn steps, so it can get very
        large.

    :Returns:

    *history* : History
        Structure containing *best*, *best_point* and *buffer*.  *best* is
        the best nllf value seen and *best_point* is the parameter vector
        which yielded *best*.  The list *buffer* contains lists of tuples
        (step, temperature, nllf, x) for each temperature.
    """
    N = len(T)
    history = History(logfile=logfile, streams=N, size=steps)
    bounder = ReflectBounds(*bounds)
    # stepper = RandStepper(bounds, tol=0.2/T[-1])
    stepper = Stepper(bounds, history)
    dT = diff(1.0 / asarray(T))
    P = asarray([p] * N)  # Points
    E = ones(N) * nllf(p)  # Values
    history.save(step=0, temperature=T, energy=E, point=P)
    total_accept = zeros(N)
    total_swap = zeros(N - 1)
    with np.errstate(over="ignore"):
        for step in range(1, steps + burn):
            # Take a step
            R = rand()
            if step < 20 or R < 0.2:
                # action = 'jiggle'
                Pnext = [stepper.jiggle(p, 0.01 * t / T[-1]) for p, t in zip(P, T)]
            elif R < 0.4:
                # action = 'direct'
                Pnext = [stepper.direct(p, i) for i, p in enumerate(P)]
            else:
                # action = 'diffev'
                Pnext = [stepper.diffev(p, i, CR=CR) for i, p in enumerate(P)]

            # Test constraints
            Pnext = asarray([bounder.apply(p) for p in Pnext])

            # Temperature dependent Metropolis update
            Enext = asarray([nllf(p) for p in Pnext])
            accept = exp(-(Enext - E) / T) > rand(N)
            # print step,action
            # print "dP"," ".join("%.6f"%norm((pn-p)/stepper.step) for pn,p in zip(P,Pnext))
            # print "dE"," ".join("%.1f"%(en-e) for en,e in zip(E,Enext))
            # print "En"," ".join("%.1f"%e for e in Enext)
            # print "accept",accept
            E[accept] = Enext[accept]
            P[accept] = Pnext[accept]
            total_accept += accept

            # Accumulate history for population based methods
            history.save(step, temperature=T, energy=E, point=P, changed=accept)
            # print "best",history.best

            # Swap chains across temperatures
            # Note that we are are shuffling from high to low so that if a good
            # point is found at a high temperature which push it immediately as
            # low as we can go rather than risk losing it at the next high temp
            # step.
            swap = zeros(N - 1)
            for i in range(N - 2, -1, -1):
                # print "swap",E[i+1]-E[i],dT[i],exp((E[i+1]-E[i])*dT[i])
                if exp((E[i + 1] - E[i]) * dT[i]) > rand():
                    swap[i] = 1
                    E[i + 1], E[i] = E[i], E[i + 1]
                    P[i + 1], P[i] = P[i] + 0, P[i + 1] + 0
            total_swap += swap
            # assert nllf(P[0]) == E[0]

            # Monitoring
            monitor(step, history.best_point, history.best, P, E)
            interval = 100
            if 0 and step % interval == 0:
                print("max r", max(["%.1f" % norm(p - P[0]) for p in P[1:]]))
                # print "min AR",argmin(total_accept),min(total_accept)
                # print "min SR",argmin(total_swap),min(total_swap)
                print("AR", total_accept)
                print("SR", total_swap)
                print("s(d)", [int(std([p[i] for p in P])) for i in (3, 7, 11, -1)])
                total_accept *= 0
                total_swap *= 0

    return history


class History(object):
    def __init__(self, streams=None, size=1000, logfile=None):
        # Allocate buffers
        self.size = size
        self.buffer = [[] for _ in range(streams)]
        # Prepare log file
        if logfile is not None:
            self.log = open(logfile, "w")
            print("# Step Temperature Energy Point", file=self.log)
        else:
            self.log = None
        # Track the optimum
        self.best = inf

    def save(self, step, temperature, energy, point, changed=None):
        if changed is None:
            changed = ones(len(temperature), "b")
        for i, a in enumerate(changed):
            if a:
                self._save_point(step, i, temperature[i], energy[i], point[i] + 0)

    def _save_point(self, step, i, T, E, P):
        # Save in buffer
        S = self.buffer[i]
        if len(S) >= self.size:
            S.pop(0)
        if len(S) > 0:
            # print "P",P
            # print "S",S[-1][3]
            assert norm(P - S[-1][3]) != 0
        S.append((step, T, E, P))
        # print "stream",i,"now len",len(S)
        # Track of the optimum
        if E < self.best:
            self.best = E
            self.best_point = P
        # Log to file
        if self.log:
            point_str = " ".join("%.6g" % v for v in P)
            print(step, T, E, point_str, file=self.log)
            self.log.flush()

    def draw(self, stream, k):
        """
        Return a list of k items drawn from the given stream.

        If the stream is too short, fewer than n items may be returned.
        """
        S = self.buffer[stream]
        n = len(S)
        return [S[i] for i in choose(n, k)] if n > k else S[:]


class Stepper(object):
    def __init__(self, bounds, history):
        low, high = bounds
        self.offset = low
        self.step = high - low
        self.history = history

    def diffev(self, p, stream, CR=0.8, noise=0.05):
        if len(self.history.buffer[stream]) < 20:
            # print "jiggling",stream,stream,len(self.history.buffer[stream])
            return self.jiggle(p, 1e-6)
        # Ideas incorporated from DREAM by Vrugt
        N = len(p)
        # Select to number of vector pair differences to use in update
        # using k ~ discrete U[1,max pairs]
        k = randint(4) + 1

        # Select 2*k members at random
        parents = [v[3] for v in self.history.draw(stream, 2 * k)]
        k = len(parents) // 2  # May not have enough parents
        pop = array(parents)
        # print "k",k
        # print "parents",parents
        # print "pop",pop

        # Select the dims to update based on the crossover ratio, making
        # sure at least one significant dim is selected
        while True:
            vars = nonzero(rand(N) < CR)
            if len(vars) == 0:
                vars = [randint(N)]
            step = np.sum(pop[:k] - pop[k:], axis=0)
            if norm(step[vars]) > 0:
                break

        # Weight the size of the jump inversely proportional to the
        # number of contributions, both from the parameters being
        # updated and from the population defining the step direction.
        gamma = 2.38 / sqrt(2 * len(vars) * k)

        # Apply that step with F scaling and noise
        eps = 1 + noise * (2 * rand(N) - 1)
        # print "j",j
        # print "gamma",gamma
        # print "step",step.shape
        # print "vars",vars.shape
        delta = zeros_like(p)
        delta[vars] = gamma * (eps * step)[vars]
        assert norm(delta) != 0
        return p + delta

    def direct(self, p, stream):
        if len(self.history.buffer[stream]) < 20:
            # print "jiggling",stream,len(self.history.buffer[stream])
            return self.jiggle(p, 1e-6)
        pair = self.history.draw(stream, 2)
        delta = pair[0][3] - pair[1][3]
        if norm(delta) == 0:
            print("direct should never return identical points!!")
            return self.random(p)
        assert norm(delta) != 0
        return p + delta

    def jiggle(self, p, noise):
        delta = randn(len(p)) * self.step * noise
        assert norm(delta) != 0
        return p + delta

    def random(self, p):
        delta = rand(len(p)) * self.step + self.offset
        assert norm(delta) != 0
        return p + delta

    def subspace_jiggle(self, p, noise, k):
        n = len(self.step)
        if n < k:
            idx = slice(None)
            k = n
        else:
            idx = choose(n, k)
        delta = zeros_like(p)
        delta[idx] = randn(k) * self.step[idx] * noise
        assert norm(delta) != 0
        return p + delta


class ReflectBounds(object):
    """
    Reflect parameter values into bounded region
    """

    def __init__(self, low, high):
        self.low, self.high = [asarray(v, "d") for v in (low, high)]

    def apply(self, y):
        """
        Update x so all values lie within bounds

        Returns x for convenience.  E.g., y = bounds.apply(x+0)
        """
        minn, maxn = self.low, self.high
        # Reflect points which are out of bounds
        idx = y < minn
        y[idx] = 2 * minn[idx] - y[idx]
        idx = y > maxn
        y[idx] = 2 * maxn[idx] - y[idx]

        # Randomize points which are still out of bounds
        idx = (y < minn) | (y > maxn)
        y[idx] = minn[idx] + rand(sum(idx)) * (maxn[idx] - minn[idx])
        return y


def choose(n, k):
    """
    Return an array of k things selected from a pool of n without replacement.
    """
    # At k == n/4, need to draw an extra 15% to get k unique draws
    if k > n / 4 or n < 100:
        idx = permutation(n)[:k]
    else:
        s = set(randint(n, size=k))
        while len(s) < k:
            s.add(randint(n))
        idx = array([si for si in s])
    if len(set(idx)) != len(idx):
        print("choose(n,k) contains dups!!", n, k)
    return idx



"""
2d array representation of a circle

References::
    None
"""

import numpy as np
from numpy import arange
from numpy import random, sin, cos, pi, inf, sqrt

random.seed(123)


def circle(coeffs, interval=0.02):
    """
    generate a 2D array representation of a circle of given coeffs
    coeffs = (x,y,r)
    """
    xc, yc, r = coeffs
    theta = arange(0, 2 * pi, interval)
    return r * cos(theta) + xc, r * sin(theta) + yc


def genpoints(coeffs, npts=20):
    """
    Generate a 2D dataset of npts enclosed in circle of given coeffs,
    where coeffs = (x,y,r).

    NOTE: if npts is None, constrain all points to circle of given radius
    """
    xo, yo, R = coeffs
    # Radial density varies as sqrt(x)
    r, theta = sqrt(random.rand(npts)), 2 * pi * (random.rand(npts))
    x, y = r * cos(theta) + xo, r * sin(theta) + yo
    return x, y


def gendensity(coeffs, density=0.1):
    xo, yo, R = coeffs
    npts = int(0.2 * pi * R**2)
    return genpoints(coeffs, npts)


from .model import Fitness


class MinimumCircle(Fitness):
    def __init__(self, data=None, limits=None, start=None):
        self.x, self.y = data
        self.dy = None
        self.limits = limits
        self.start = start

    def _residuals(self, p):
        x, y = self.x, self.y
        xc, yc, r = p
        d = sqrt((x - xc) ** 2 + (y - yc**2))
        d[d < r] = 0
        return d

    def profile(self, p):
        return circle(p)

    def residuals(self, p):
        """
        Residuals is used by Levenburg-Marquardt, so fake the
        penalty terms of the normal cost function for use here.
        """
        resid = self._residuals(p)
        # Throw r in the residual so that it is minimized, punish the circle
        # if there are too many points outside.
        d = np.concatenate((resid, self.r, sum(resid > 0)))
        return d

    def __call__(self, p):
        xc, yc, r = p
        resid = self._residuals(p)
        # Penalties are the number
        # Add additional penalties for each point outside the circle
        return sum(resid > 0) + sum(resid) + abs(r)


# prepared instances
Po = [0, 0, 1]
Plo = [-inf, -inf, 0]
Phi = [inf, inf, inf]
dense_circle = MinimumCircle(data=gendensity([3, 4, 20], density=5), limits=(Plo, Phi), start=Po)
sparse_circle = MinimumCircle(data=gendensity([3, -4, 32], density=0.5), limits=(Plo, Phi), start=Po)
minimal_circle = MinimumCircle(data=gendensity([0, 0, 10], density=5), limits=(Plo, Phi), start=Po)



from copy import deepcopy
from threading import Thread
import os

import wx.lib.newevent

from .. import monitor
from ..fitters import FitDriver
from ..mapper import MPMapper, SerialMapper, can_pickle
from ..util import redirect_console

from .convergence_view import ConvergenceMonitor
# ==============================================================================

PROGRESS_DELAY = 5
IMPROVEMENT_DELAY = 5

(FitProgressEvent, EVT_FIT_PROGRESS) = wx.lib.newevent.NewEvent()
(FitCompleteEvent, EVT_FIT_COMPLETE) = wx.lib.newevent.NewEvent()


# NOTE: GUI monitors are running in a separate thread.  They should not
# touch the problem internals.
class GUIProgressMonitor(monitor.TimedUpdate):
    def __init__(self, win, problem, progress=None, improvement=None):
        monitor.TimedUpdate.__init__(
            self, progress=progress or PROGRESS_DELAY, improvement=improvement or IMPROVEMENT_DELAY
        )
        self.win = win
        self.problem = problem

    def show_progress(self, history):
        evt = FitProgressEvent(
            problem=self.problem,
            message="progress",
            step=history.step[0],
            value=history.value[0],
            point=history.point[0] + 0,
        )  # avoid race
        wx.PostEvent(self.win, evt)

    def show_improvement(self, history):
        evt = FitProgressEvent(
            problem=self.problem,
            message="improvement",
            step=history.step[0],
            value=history.value[0],
            point=history.point[0] + 0,
        )  # avoid race
        wx.PostEvent(self.win, evt)


class GUIMonitor(monitor.Monitor):
    """
    Generic GUI monitor for fitting.

    Sends a fit progress event to the selected window every *rate*
    seconds. The *monitor* is a bumps monitor which processes fit
    history as it arrives. Instead of the usual *show_progress* method
    to update the managing process it has a *progress* method which
    returns a dictionary of progress attributes.  These attributes are
    added to a *FitProgressEvent* along with *problem* and *message* which
    is then posted to *win*.

    *problem* should be the fit problem handed to the fit thread, and not
    a copy. This is because it is used for direct comparison with the current
    fit object in the progress panels so that stray messages don't cause
    confusion when making graphs.

    *message* is a dispatch string used by the OnFitProgress event processor
    in the app to determine which progress panel should receive the event.

    TODO: this is a pretty silly design, especially since GUI monitor is only
    used for the convergence plot, and a separate monitor class was created
    for DREAM progress updates.
    """

    def __init__(self, win, problem, message, monitor, rate=None):
        self.time = 0
        self.rate = rate  # rate=0 for no progress update, only final
        self.win = win
        self.problem = problem
        self.message = message
        self.monitor = monitor

    def config_history(self, history):
        self.monitor.config_history(history)
        history.requires(time=1)

    def __call__(self, history):
        self.monitor(history)
        if self.rate > 0 and history.time[0] >= self.time + self.rate:
            evt = FitProgressEvent(problem=self.problem, message=self.message, **self.monitor.progress())
            wx.PostEvent(self.win, evt)
            self.time = history.time[0]

    def final(self):
        """
        Close out the monitor
        """
        evt = FitProgressEvent(problem=self.problem, message=self.message, **self.monitor.progress())
        wx.PostEvent(self.win, evt)


# Horrible hacks:
# (1) We are grabbing uncertainty_state from the history on monitor update
# and holding onto it for the monitor final call. Need to restructure the
# history/monitor interaction so that object lifetimes are better controlled.
# (2) We set the uncertainty state directly in the GUI window. This is an
# attempt to work around a memory leak which causes problems in the GUI when
# it is updated too frequently. If the problem is that the event structure
# is too large and wx isn't cleaning up properly then this should address it,
# but if the problem lies within the DREAM plot functions or within matplotlib
# then this change won't do anything and should be reverted.
class DreamMonitor(monitor.Monitor):
    def __init__(self, win, problem, message, fitter, rate=None):
        self.time = 0
        self.rate = rate  # rate=0 for no progress update, only final
        self.win = win
        self.problem = problem
        self.fitter = fitter
        self.message = message
        self.uncertainty_state = None

    def config_history(self, history):
        history.requires(time=1)

    def __call__(self, history):
        self.uncertainty_state = getattr(history, "uncertainty_state", None)
        if self.rate > 0 and history.time[0] >= self.time + self.rate and self.uncertainty_state is not None:
            # Note: win.uncertainty_state protected by win.fit_lock
            self.time = history.time[0]
            self.win.uncertainty_state = self.uncertainty_state
            evt = FitProgressEvent(
                problem=self.problem,
                message="uncertainty_update",
                # uncertainty_state=deepcopy(self.uncertainty_state),
            )
            wx.PostEvent(self.win, evt)

    def final(self):
        """
        Close out the monitor
        """
        if self.uncertainty_state is not None:
            # Note: win.uncertainty_state protected by win.fit_lock
            self.win.uncertainty_state = self.uncertainty_state
            evt = FitProgressEvent(
                problem=self.problem,
                message="uncertainty_final",
                # uncertainty_state=deepcopy(self.uncertainty_state),
            )
            wx.PostEvent(self.win, evt)


# ==============================================================================


class FitThread(Thread):
    """Run the fit in a separate thread from the GUI thread."""

    def __init__(
        self,
        win,
        abort_test=None,
        problem=None,
        fitclass=None,
        options=None,
        mapper=None,
        convergence_update=5,
        uncertainty_update=300,
    ):
        # base class initialization
        # Process.__init__(self)

        Thread.__init__(self)
        self.win = win
        self.abort_test = abort_test
        self.problem = problem
        self.fitclass = fitclass
        self.options = options
        self.mapper = mapper
        self.convergence_update = convergence_update
        self.uncertainty_update = uncertainty_update

    def run(self):
        # TODO: we have no interlocks on changes in problem state.  What
        # happens when the user changes the problem while a fit is being run?
        # May want to keep a history of changes to the problem definition,
        # along with a function to reverse them so we can handle undo.

        # NOTE: Problem must be the original problem (not a copy) when used
        # inside the GUI monitor otherwise AppPanel will not be able to
        # recognize that it is the same problem when updating views.
        monitors = [
            GUIProgressMonitor(self.win, self.problem),
            GUIMonitor(
                self.win,
                self.problem,
                message="convergence_update",
                monitor=ConvergenceMonitor(),
                rate=self.convergence_update,
            ),
            DreamMonitor(
                self.win, self.problem, fitter=self.fitclass, message="uncertainty_update", rate=self.uncertainty_update
            ),
        ]
        # Only use parallel if the problem can be pickled
        mapper = MPMapper if can_pickle(self.problem) else SerialMapper

        # Be safe and send a private copy of the problem to the fitting engine
        # print "fitclass",self.fitclass
        problem = deepcopy(self.problem)
        # print "fitclass id",id(self.fitclass),self.fitclass,threading.current_thread()
        driver = FitDriver(
            self.fitclass,
            problem=problem,
            monitors=monitors,
            abort_test=self.abort_test,
            mapper=mapper.start_mapper(problem, []),
            **self.options,
        )

        x, fx = driver.fit()
        # Give final state message from monitors
        for M in monitors:
            if hasattr(M, "final"):
                M.final()

        with redirect_console() as fid:
            driver.show()
            captured_output = fid.getvalue()

        evt = FitCompleteEvent(problem=self.problem, point=x, value=fx, info=captured_output)
        wx.PostEvent(self.win, evt)



"""
Parallel and serial mapper implementations.

The API is a bit crufty since interprocess communication has evolved from
the original implementation. And the names are misleading.

Usage::

    Mapper.start_worker(problem)
    mapper = Mapper.start_mapper(problem, None, cpus)
    result = mapper(points)
    ...
    mapper = Mapper.start_mapper(problem, None, cpus)
    result = mapper(points)
    Mapper.stop_mapper(mapper)
"""

import sys
import os
import signal

# {{{ http://code.activestate.com/recipes/496767/ (r1)
# Converted to use ctypes by Paul Kienzle


PROCESS_ALL_ACCESS = 0x1F0FFF


def can_pickle(problem, check=False):
    """
    Returns True if *problem* can be pickled.

    If this method returns False then MPMapper cannot be used and
    SerialMapper should be used instead.

    If *check* is True then call *nllf()* on the duplicated object as a
    "smoke test" to verify that the function will run after copying. This
    is not foolproof. For example, access to a database may work in the
    duplicated object because the connection is open and available in the
    current process, but it will fail when trying to run on a remote machine.
    """
    try:
        import dill
    except ImportError:
        dill = None
        import pickle
    try:
        if dill is not None:
            dup = dill.loads(dill.dumps(problem, recurse=True))
        else:
            dup = pickle.loads(pickle.dumps(problem))
        if check:
            dup.nllf()
        return True
    except Exception:
        return False


def setpriority(pid=None, priority=1):
    """
    Set The Priority of a Windows Process.  Priority is a value between 0-5
    where 2 is normal priority and 5 is maximum.  Default sets the priority
    of the current python process but can take any valid process ID.
    """

    # import win32api,win32process,win32con
    from ctypes import windll

    priorityclasses = [
        0x40,  # IDLE_PRIORITY_CLASS,
        0x4000,  # BELOW_NORMAL_PRIORITY_CLASS,
        0x20,  # NORMAL_PRIORITY_CLASS,
        0x8000,  # ABOVE_NORMAL_PRIORITY_CLASS,
        0x80,  # HIGH_PRIORITY_CLASS,
        0x100,  # REALTIME_PRIORITY_CLASS
    ]
    if pid is None:
        pid = windll.kernel32.GetCurrentProcessId()
    handle = windll.kernel32.OpenProcess(PROCESS_ALL_ACCESS, True, pid)
    windll.kernel32.SetPriorityClass(handle, priorityclasses[priority])


# end of http://code.activestate.com/recipes/496767/ }}}


def nice():
    if os.name == "nt":
        setpriority(priority=1)
    else:
        os.nice(5)


class SerialMapper(object):
    @staticmethod
    def start_worker(problem):
        pass

    @staticmethod
    def start_mapper(problem, modelargs=None, cpus=0):
        # Note: map is n iterator in python 3.x
        return lambda points: list(map(problem.nllf, points))

    @staticmethod
    def stop_mapper(mapper):
        pass


# Load the problem in the remote process rather than pickling
# def _MP_load_problem(*modelargs):
#    from .fitproblem import load_problem
#    _MP_set_problem(load_problem(*modelargs))


def _MP_setup(namespace):
    # Using MPMapper class variables to store worker globals.
    # It doesn't matter if they conflict with the controller values since
    # they are in a different process.
    signal.signal(signal.SIGINT, signal.SIG_IGN)
    MPMapper.namespace = namespace
    nice()


def _MP_run_problem(problem_point_pair):
    problem_id, point = problem_point_pair
    if problem_id != MPMapper.problem_id:
        # print(f"Fetching problem {problem_id} from namespace")
        # Problem is pickled using dill when it is available
        try:
            import dill

            MPMapper.problem = dill.loads(MPMapper.namespace.pickled_problem)
        except ImportError:
            MPMapper.problem = MPMapper.namespace.problem
        MPMapper.problem_id = problem_id
    return MPMapper.problem.nllf(point)


class MPMapper(object):
    # Note: suprocesses are using the same variables
    pool = None
    manager = None
    namespace = None
    problem_id = 0

    @staticmethod
    def start_worker(problem):
        pass

    @staticmethod
    def start_mapper(problem, modelargs=None, cpus=0):
        import multiprocessing

        # Set up the process pool on the first call.
        if MPMapper.pool is None:
            # Create a sync namespace to distribute the problem description.
            MPMapper.manager = multiprocessing.Manager()
            MPMapper.namespace = MPMapper.manager.Namespace()
            # Start the process pool, sending the namespace handle
            if cpus == 0:
                cpus = multiprocessing.cpu_count()
            MPMapper.pool = multiprocessing.Pool(cpus, _MP_setup, (MPMapper.namespace,))

        # Increment the problem number and store the problem in the namespace.
        # The store action uses pickle to transfer python objects to the
        # manager process. Since this may fail for lambdas and for functions
        # defined within the model file, instead use dill (if available)
        # to pickle the problem before storing.
        MPMapper.problem_id += 1
        try:
            import dill

            MPMapper.namespace.pickled_problem = dill.dumps(problem, recurse=True)
        except ImportError:
            MPMapper.namespace.problem = problem
        ## Store the modelargs and the problem name if pickling doesn't work
        # MPMapper.namespace.modelargs = modelargs

        # Set the mapper to send problem_id/point value pairs
        def mapper(points):
            try:
                return MPMapper.pool.map(_MP_run_problem, ((MPMapper.problem_id, p) for p in points))
            except KeyboardInterrupt:
                MPMapper.stop_mapper(None)

        return mapper

    @staticmethod
    def stop_mapper(mapper):
        # reset pool and manager
        MPMapper.pool.terminate()
        MPMapper.manager.shutdown()
        MPMapper.pool = None
        MPMapper.manager = None
        MPMapper.namespace = None
        # Don't reset problem id; it keeps count even when mapper is restarted.
        ##MPMapper.problem_id = 0


def _MPI_set_problem(problem, comm, root=0):
    import dill

    pickled_problem = dill.dumps(problem, recurse=True) if comm.rank == root else None
    pickled_problem = comm.bcast(pickled_problem, root=root)
    return problem if comm.rank == root else dill.loads(pickled_problem)


def _MPI_map(problem, points, comm, root=0):
    # print(f"{comm.rank}: mapping points")
    import numpy as np
    from mpi4py import MPI

    # Send number of points and number of variables per point.
    # root: return result if there are points otherwise return False
    # worker: return True if there are points otherwise return False
    npoints, nvars = comm.bcast(points.shape if comm.rank == root else None, root=root)
    if npoints == 0:
        return False

    # Divvy points equally across all processes
    whole = points if comm.rank == root else None
    idx = np.arange(comm.size)
    size = np.ones(comm.size, idx.dtype) * (npoints // comm.size) + (idx < npoints % comm.size)
    offset = np.cumsum(np.hstack((0, size[:-1])))
    part = np.empty((size[comm.rank], nvars), dtype="d")
    comm.Scatterv((whole, (size * nvars, offset * nvars), MPI.DOUBLE), (part, MPI.DOUBLE), root=root)

    # Evaluate models assigned to each processor
    partial_result = np.array([problem.nllf(pk) for pk in part], dtype="d")

    # Collect results
    result = np.empty(npoints, dtype="d") if comm.rank == root else True
    comm.Barrier()
    comm.Gatherv((partial_result, MPI.DOUBLE), (result, (size, offset), MPI.DOUBLE), root=root)
    comm.Barrier()
    return result


class MPIMapper(object):
    _first_fit = True  # The first problem is set when the worker starts

    @staticmethod
    def start_worker(problem):
        """
        Start the worker process.

        For the main process this does nothing and returns immediately. The
        worker processes never return.

        Each worker sits in a loop waiting for the next batch of points
        for the problem, or for the next problem. Set t
        problem is set to None, then exit the process and never
        """
        from mpi4py import MPI

        comm, root = MPI.COMM_WORLD, 0

        # If worker, sit in a loop waiting for the next point.
        # If the point is empty, then wait for a new problem.
        # If the problem is None then we are done, otherwise wait for next point.
        if comm.rank != root:
            # print(f"{comm.rank}: looping")
            while True:
                result = _MPI_map(problem, None, comm, root)
                if not result:
                    problem = _MPI_set_problem(None, comm, root)
                    if problem is None:
                        break
                    # print(f"{comm.rank}: changing problem")

            # print(f"{comm.rank}: finalizing")
            MPI.Finalize()

            # Exit the program after the worker is done. Don't return
            # to the caller since that is continuing on with the main
            # thread, and in particular, attempting to rerun the fit on
            # each worker.
            sys.exit(0)

    @staticmethod
    def start_mapper(problem, modelargs=None, cpus=0):
        # Only root can get here---worker is stuck in start_worker
        from mpi4py import MPI

        comm, root = MPI.COMM_WORLD, 0
        import numpy as np

        # Signal new problem then send it, but not on the first fit. We do this
        # so that we can still run MPI fits even if the problem itself cannot
        # be pickled, but only the first one. (You can still fit a series even
        # if the problem can't be pickled, but you will need to restart the
        # MPI job separately for each fit.)
        # Note: setting problem to None stops the program, so call finalize().
        mapper = lambda points: _MPI_map(problem, points, comm, root)
        if not MPIMapper._first_fit:
            # print(f"{comm.rank}: replacing problem")
            # Send an empty set of points to signal a new problem is coming.
            mapper(np.empty((0, 0), "d"))
            _MPI_set_problem(problem, comm, root)
            if problem is None:
                # print(f"{comm.rank}: finalizing root")
                MPI.Finalize()
        MPIMapper._first_fit = False
        return mapper

    @staticmethod
    def stop_mapper(mapper):
        # Set problem=None to stop the program.
        MPIMapper.start_mapper(None, None)


class AMQPMapper(object):
    @staticmethod
    def start_worker(problem):
        # sys.stderr = open("bumps-%d.log"%os.getpid(),"w")
        # print >>sys.stderr,"worker is starting"; sys.stdout.flush()
        from amqp_map.config import SERVICE_HOST
        from amqp_map.core import connect, start_worker as serve

        server = connect(SERVICE_HOST)
        # os.system("echo 'serving' > /tmp/map.%d"%(os.getpid()))
        # print "worker is serving"; sys.stdout.flush()
        serve(server, "bumps", problem.nllf)
        # print >>sys.stderr,"worker ended"; sys.stdout.flush()

    @staticmethod
    def start_mapper(problem, modelargs=None, cpus=0):
        import sys
        import multiprocessing
        import subprocess
        from amqp_map.config import SERVICE_HOST
        from amqp_map.core import connect, Mapper

        server = connect(SERVICE_HOST)
        mapper = Mapper(server, "bumps")
        cpus = multiprocessing.cpu_count()
        pipes = []
        for _ in range(cpus):
            cmd = [sys.argv[0], "--worker"] + modelargs
            # print "starting",sys.argv[0],"in",os.getcwd(),"with",cmd
            pipe = subprocess.Popen(cmd, universal_newlines=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
            pipes.append(pipe)
        for pipe in pipes:
            if pipe.poll() > 0:
                raise RuntimeError(
                    "subprocess returned %d\nout: %s\nerr: %s" % (pipe.returncode, pipe.stdout, pipe.stderr)
                )
        # os.system(" ".join(cmd+["&"]))
        import atexit

        def exit_fun():
            for p in pipes:
                p.terminate()

        atexit.register(exit_fun)

        # print "returning mapper",mapper
        return mapper

    @staticmethod
    def stop_mapper(mapper):
        for pipe in mapper.pipes:
            pipe.terminate()



# persist settings to disk

import sys
from pathlib import Path
from typing import Optional

ETC_PATH = Path(sys.prefix, "etc")
DEFAULT_APPLICATION = "bumps"


def set_value(key: str, value: str, application: str = DEFAULT_APPLICATION):
    settings_path = ETC_PATH / application / key
    try:
        settings_path.parent.mkdir(parents=True, exist_ok=True)
        with open(settings_path, "w") as f:
            f.write(value)
        return True
    except Exception as e:
        return False


def get_value(key: str, default_value: Optional[str] = None, application: str = DEFAULT_APPLICATION):
    settings_path = ETC_PATH / application / key
    try:
        with open(settings_path, "r") as f:
            return f.read()
    except Exception as e:
        return default_value



# This code is public domain
# Author: Paul Kienzle

## Differential Evolution Solver Class
## Based on algorithms developed by Dr. Rainer Storn & Kenneth Price
## Influenced by
##    Lester E. Godwin (godwin@pushcopr.com)  1998: C++ version
##    James R. Phillips (zunzun@zunzun.com)   2002: Python conversion
##    Patrick Hung                            2006: cleanup
##    Mike McKerns (mmckerns@caltech.edu)     2008: parallel version, bounds
##    Paul Kienzle                            2009: rewrite

"""
Differential evolution optimizer.

This module contains a collection of optimization routines based on
Storn and Price's differential evolution algorithm.

Minimal function interface to optimization routines::

    x = diffev(f,xo)

Stepping interface::

    DifferentialEvolution


References
==========

[1] Storn, R. and Price, K. Differential Evolution - A Simple and Efficient
Heuristic for Global Optimization over Continuous Spaces. Journal of Global
Optimization 11: 341-359, 1997.

[2] Price, K., Storn, R., and Lampinen, J. - Differential Evolution,
A Practical Approach to Global Optimization. Springer, 1st Edition, 2005

"""

# Symbols required for simple interface
__all__ = ["de", "stop"]

import numpy as np

from .. import stop
from .. import solver

CROSSOVER = "c_exp", "c_bin"
MUTATE = "best1", "best1u", "best2", "randtobest1", "rand1", "rand2"


#############################################################################
#  Code below are the different crossovers/mutation strategies
#############################################################################
def c_exp(ndim, CR):
    """
    Select a sequence of dimensions.

    The length of the sequence follows the geometric distribution for 1-CR.

    This is equivalent to flipping the first heads after n flips with
    a weighted coin.

    The sequence starts at a random position and wraps if necessary.
    """
    # Note: this is different from Patrick Hung's version in that it forces
    # at least one success.
    L = min(abs(np.random.geometric(1 - CR)), ndim)
    idx = np.zeros(ndim, "bool")
    n = np.random.randint(ndim)
    idx[np.arange(n, n + L) % ndim] = True
    return idx


def c_bin(ndim, CR):
    """
    Select random dimensions.

    The probability of selecting any dimension is CR.  At least one dimension
    will be selected.
    """
    n = np.random.randint(ndim)
    idx = np.random.rand(ndim) < CR
    idx[n] = True
    return idx


def best1(F, best, pop, idx, dims):
    """
    Differential evolution mutation T = best + F*(r1-r2)
    """
    r1, r2 = _candidates(pop, 2, exclude=idx)
    return best[dims] + F * (r1[dims] - r2[dims])


def best1u(F, best, pop, idx, dims):
    """
    Differential evolution mutation T = best + U*(r1-r2),  U ~ Uniform[0,F]
    """
    r1, r2 = _candidates(pop, 2, exclude=idx)
    return best[dims] + F * np.random.rand() * (r1[dims] - r2[dims])


def best2(F, best, pop, idx, dims):
    """
    Differential evolution mutation T = best + F*(r1+r2-r3-r4)
    """
    r1, r2, r3, r4 = _candidates(pop, 4, exclude=idx)
    return best[dims] + F * (r1[dims] + r2[dims] - r3[dims] - r4[dims])


def randtobest1(F, best, pop, idx, dims):
    """
    Differential evolution mutation T = F*(best-old + r1-r2)
    """
    r1, r2 = _candidates(pop, 2, exclude=idx)
    return F * (best[dims] - pop[idx][dims] + r1[dims] - r2[dims])


def rand1(F, best, pop, idx, dims):
    """
    Differential evolution mutation T = r0 + F*(r1-r2)
    """
    r0, r1, r2 = _candidates(pop, 3, exclude=idx)
    return r0[dims] + F * (r1[dims] - r2[dims])


def rand1u(F, best, pop, idx, dims):
    """
    Differential evolution mutation T = r0 + U*(r1-r2), U ~ Uniform[0,F]
    """
    r0, r1, r2 = _candidates(pop, 3, exclude=idx)
    return r0[dims] + F * np.random.rand() * (r1[dims] - r2[dims])


def rand2(F, best, pop, idx, dims):
    """
    Differential evolution mutation T = r0 + F*(r1+r2-r3-r4)
    """
    r0, r1, r2, r3, r4 = _candidates(pop, 5, exclude=idx)
    return r0[dims] + F * (r1[dims] + r2[dims] - r3[dims] - r4[dims])


############################################################


def _candidates(pop, k, exclude=None):
    """
    Select *n* random candidates from *pop*, not including the
    candidate at index *exclude*.
    """
    n = len(pop)
    if exclude is not None:
        selection = np.arange(n - 1, dtype="i")
        if exclude < n - 1:
            selection[exclude] = n - 1
    else:
        selection = np.arange(n, dtype="i")
    np.random.shuffle(selection)
    return pop[selection[:k]]


##########################################################################


class DifferentialEvolution(solver.Strategy):
    """
    Differential evolution optimization.

    *CR*  float in [0-1]
        Crossover rate.
    *F* float in (0,inf)
        Crossover step size.
    *npop* float
        The size of the population is npop times the number of dimensions
        in the problem.
    *crossover* func(ndim, CR) -> index vector
        Crossover selection.  Returns the index vector of dimensions which
        should be mutated.
    *mutate* fn(F, best, pop, idx, dims) -> new[dims]
        Mutation strategy.  Selects the crossover population members and
        returns the mutated portion of the trial point in the new population.
        *F* is the scale factor, *best* is the best point seen so far, *pop*
        is the current population, *idx* is the vector being updated and
        *dims* is the set of dimensions to update.

    Available crossover functions (block is default)::

        c_exp:  start at dimension n and continue until U[0,1] >= CR
        c_bin:  select dimension i if U[0,1] >= CR

    Available mutation functions (best1u is default)::

        best1u: T = best + U(F)*(r1-r2),  U(F) ~ Uniform in [0,F]
        best1:  T = best + F*(r1-r2)
        best2:  T = best + F*(r1+r2-r3-r4)
        randtobest1:  T = F*(best-old) + F*(r1-r2)
        rand1:  T = r0 + F*(r1-r2)
        rand2:  T = r0 + F*(r1+r2-r3-r4)
    """

    requires = [("mystic", "0.9")]

    def __init__(self, CR=0.5, F=2.0, npop=3, crossover=c_exp, mutate=best1u):
        self.crossover = crossover
        self.mutate = mutate
        self.CR, self.F = CR, F
        self.npop = npop

    def default_termination_conditions(self, problem):
        success = stop.Cf(tol=1e-7, scaled=False)
        # maxiter = 100
        maxiter = len(problem.getp()) * 200
        # maxfun  = self.npop*maxiter
        failure = stop.Steps(maxiter)
        return success, failure

    def config_history(self, history):
        """
        Indicates how much history is required.
        """
        history.requires(value=1, population_points=2, population_values=2)

    def start(self, problem):
        """
        Generate the initial population.

        Returns a matrix *P* of points to be evaluated.
        """
        # Generate a random population
        current = problem.getp()
        ndim = len(current)
        population = problem.randomize(int(self.npop * ndim))
        population[0] = current

        # Return the population
        return population

    def step(self, history):
        """
        Generate the next population.

        Returns a matrix *P* of points to be evaluated.

        *history* contains the previous history of the computation,
        including any fields placed by :meth:`update`.
        """

        best = history.point[0]
        pop = history.population_points[0]
        pop_size, ndim = pop.shape

        trial = pop.copy()
        for idx, vec in enumerate(trial):
            dims = self.crossover(ndim, self.CR)
            vec[dims] = self.mutate(self.F, best, pop, idx, dims)
        return trial

    def update(self, history):
        """
        Update population, keeping old points that are better than
        the trial points.
        """
        # print "result",history.step[0]
        # for i,v in enumerate(history.population_values[0]):
        #    print history.population_points[0][i],'=',v
        if len(history.population_points) > 1:
            oldpop = history.population_points[1]
            oldval = history.population_values[1]
            newpop = history.population_points[0]
            newval = history.population_values[0]

            worse = newval > oldval
            newpop[worse] = oldpop[worse]
            newval[worse] = oldval[worse]


# minimizer_function(strategy=DifferentialEvolution,
#                   success=stop.Df(1e-5,n=10),
#                   failure=stop.Steps(100))



# Copyright (C) 2006-2011, University of Maryland
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/ or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# Author: James Krycka

"""
This module implements the AppFrame class which creates the main frame of the
GUI for the Bumps application including a basic menu, tool bar, and status bar.
"""

# ==============================================================================

import sys

import wx

from .about import AboutDialog, APP_TITLE, APP_DESCRIPTION, APP_LICENSE, APP_CREDITS, APP_TUTORIAL
from .app_panel import AppPanel
from .console import NumpyConsole
from .utilities import resource, choose_fontsize, display_fontsize

# Resource files.
PROG_ICON = "bumps.ico"


# ==============================================================================
class ModelConsole(NumpyConsole):
    def OnChanged(self, added=(), changed=(), removed=()):
        pass

    def OnClose(self, event):
        self.Show(False)


class AppFrame(wx.Frame):
    """
    This class creates the top-level frame for the application and populates it
    with application specific panels and widgets.
    """

    def __init__(
        self, parent=None, id=wx.ID_ANY, title=APP_TITLE, pos=wx.DefaultPosition, size=wx.DefaultSize, name="AppFrame"
    ):
        wx.Frame.__init__(self, parent, id, title, pos, size, name=name)

        # Display the application's icon in the title bar.
        icon = wx.Icon(resource(PROG_ICON), wx.BITMAP_TYPE_ICO)
        self.SetIcon(icon)

        # Set the default font family and font size for the application.
        self.set_default_font()

        # Initialize the menu bar with common items.
        self.add_menubar()

        # Initialize the tool bar.
        self.add_toolbar()

        # Initialize the status bar.
        self.add_statusbar()

        # Build the application panels for the GUI on the frame.
        self.panel = AppPanel(self)
        self.panel.console = ModelConsole(self)
        self.panel.console["app"] = self

        # Note: Do not call self.Fit() as this will reduce the frame to its
        # bare minimum size; we want it to keep its default size.

    def set_default_font(self):
        """
        Sets the default font family and font size for the frame which will be
        inherited by all child windows subsequently created.
        """

        # Save the system default font information before we make any changes.
        default_fontname = self.GetFont().GetFaceName()
        default_fontsize = self.GetFont().GetPointSize()

        # If requested, override the font name to use.  Note that:
        # - the MS Windows default font appears to be the same as Tahoma
        # - Arial tends to be narrower and taller than Tahoma.
        # - Verdana tends to be wider and shorter than Tahoma.
        fontname = default_fontname
        if len(sys.argv) > 1:
            if "--tahoma" in sys.argv[1:]:
                fontname = "Tahoma"
            if "--arial" in sys.argv[1:]:
                fontname = "Arial"
            if "--verdana" in sys.argv[1:]:
                fontname = "Verdana"

        fontsize = choose_fontsize(fontname=fontname)

        # If requested, override the font point size to use.
        if len(sys.argv) > 1:
            if "--12pt" in sys.argv[1:]:
                fontsize = 12
            if "--11pt" in sys.argv[1:]:
                fontsize = 11
            if "--10pt" in sys.argv[1:]:
                fontsize = 10
            if "--9pt" in sys.argv[1:]:
                fontsize = 9
            if "--8pt" in sys.argv[1:]:
                fontsize = 8
            if "--7pt" in sys.argv[1:]:
                fontsize = 7
            if "--6pt" in sys.argv[1:]:
                fontsize = 6

        # Set the default font for this and all child windows.  The font of the
        # frame's title bar is not affected (which is a good thing).  However,
        # setting the default font does not affect the font used in the frame's
        # menu bar or menu items (which is not such a good thing because the
        # menu text size be different than the size used by the application's
        # other widgets).  The menu font cannot be changed by wxPython.
        self.SetFont(wx.Font(fontsize, wx.SWISS, wx.NORMAL, wx.NORMAL, False, fontname))

        # If requested, display font and miscellaneous platform information.
        if len(sys.argv) > 1 and "--platform" in sys.argv[1:]:
            print("*** Platform =", wx.PlatformInfo)
            print("*** Default font is %s  Chosen font is %s" % (default_fontname, self.GetFont().GetFaceName()))
            print(
                "*** Default point size = %d  Chosen point size = %d"
                % (default_fontsize, self.GetFont().GetPointSize())
            )
            display_fontsize(fontname=fontname)

    def add_menubar(self):
        """Creates a default menu bar, menus, and menu options."""

        # Create the menu bar.
        mb = wx.MenuBar()
        # wx.MenuBar.SetAutoWindowMenu(False)

        # Add a 'File' menu to the menu bar and define its options.
        file_menu = wx.Menu()

        _item = file_menu.Append(wx.ID_ANY, "&Exit", "Terminate application")
        self.Bind(wx.EVT_MENU, self.OnExit, _item)

        mb.Append(file_menu, "&File")

        # Add a 'Help' menu to the menu bar and define its options.
        help_menu = wx.Menu()

        _item = help_menu.Append(wx.ID_ANY, "&About", "Get description of application")
        self.Bind(wx.EVT_MENU, self.OnAbout, _item)
        _item = help_menu.Append(wx.ID_ANY, "&Documentation", "Get User's Guide and Reference Manual")
        self.Bind(wx.EVT_MENU, self.OnTutorial, _item)
        _item = help_menu.Append(wx.ID_ANY, "License", "Read license and copyright notice")
        self.Bind(wx.EVT_MENU, self.OnLicense, _item)
        _item = help_menu.Append(wx.ID_ANY, "Credits", "Get list of authors and sponsors")
        self.Bind(wx.EVT_MENU, self.OnCredits, _item)

        help_menu.AppendSeparator()
        _item = help_menu.Append(wx.ID_ANY, "&Console", "Interactive Python shell")
        self.Bind(wx.EVT_MENU, self.OnConsole, _item)

        mb.Append(help_menu, "&Help")

        # Attach the menu bar to the frame.
        self.SetMenuBar(mb)

    def add_toolbar(self):
        """Creates a default tool bar."""

        # tb = self.CreateToolBar()
        tb = wx.ToolBar(parent=self, style=wx.TB_HORIZONTAL | wx.NO_BORDER)
        tb.Realize()
        self.SetToolBar(tb)

    def add_statusbar(self):
        """Creates a default status bar."""

        sb = self.statusbar = self.CreateStatusBar()
        sb.SetFieldsCount(1)

    def OnAbout(self, evt):
        """Shows the About dialog box."""

        dlg = AboutDialog(
            parent=self,
            title="About",
            info=APP_DESCRIPTION,
            show_name=True,
            show_notice=True,
            show_link=True,
            show_link_docs=True,
        )
        dlg.ShowModal()
        dlg.Destroy()

    def OnCredits(self, evt):
        """Shows the Credits dialog box."""

        dlg = AboutDialog(
            parent=self,
            title="Credits",
            info=APP_CREDITS,
            show_name=True,
            show_notice=True,
            show_link=False,
            show_link_docs=False,
        )
        dlg.ShowModal()
        dlg.Destroy()

    def OnExit(self, event):
        """Terminates the program."""
        self.Close()

    def OnLicense(self, evt):
        """Shows the License dialog box."""

        dlg = AboutDialog(
            parent=self,
            title="License",
            info=APP_LICENSE,
            show_name=True,
            show_notice=True,
            show_link=False,
            show_link_docs=False,
        )
        dlg.ShowModal()
        dlg.Destroy()

    def OnTutorial(self, event):
        """Shows the Tutorial dialog box."""

        dlg = AboutDialog(
            parent=self,
            title="Tutorial",
            info=APP_TUTORIAL,
            show_name=False,
            show_notice=False,
            show_link=False,
            show_link_docs=True,
        )
        dlg.ShowModal()
        dlg.Destroy()

    def OnConsole(self, event):
        """Raise python console."""
        self.panel.console.Show(True)



"""
Build layout for histogram plots (plotly)
"""

__all__ = ["var_plot_size", "plot_vars", "plot_var"]

from math import ceil, sqrt
import numpy as np

import typing

if typing.TYPE_CHECKING:
    from bumps.dream.state import Draw
    from bumps.dream.stats import VarStats
    import plotly.graph_objects as go

# TODO: improve type hinting on VarStats (dataclass with optional fields?)

# # Set space between plots in horiz and vert.
# H_SPACE = 0.2
# V_SPACE = 0.2

# # Set top, bottom, left margins.
# T_MARGIN = 0.2
# B_MARGIN = 0.2
# L_MARGIN = 0.2
# R_MARGIN = 0.4

# # Set desired plot sizes.
# TILE_W = 3.0
# TILE_H = 2.0
# CBAR_WIDTH = 0.75

CBAR_COLORS = 64
LINE_COLOR = "red"
ANNOTATION_COLOR = "blue"
HISTOGRAM_COLORMAP = "Cividis"


def tile_axes_square(n):
    """
    Determine number of columns by finding the
    next greatest square, then determine number
    of rows needed.
    """
    cols = int(ceil(sqrt(n)))
    rows = int(ceil(n / float(cols)))
    return cols, rows


def plot_vars(draw: "Draw", all_vstats, cbar_colors=CBAR_COLORS, **kw):
    from plotly.subplots import make_subplots

    n = len(all_vstats)
    ncol, nrow = tile_axes_square(n)

    snllf = np.sort(-draw.logp)
    vmin, vmax = snllf[0], snllf[int(0.98 * (len(snllf) - 1))]  # robust range
    cbar_edges = np.linspace(vmin, vmax, cbar_colors)
    # titles = [vstats.label for vstats in all_vstats]

    # fig = make_subplots(rows=nrow, cols=ncol, subplot_titles=titles)
    fig = make_subplots(rows=nrow, cols=ncol)
    fig.update_yaxes(secondary_y=True)
    fig.update_xaxes(dict(exponentformat="e"), overwrite=True)
    fig.update_xaxes(showticklabels=True, showline=True, mirror=True, linewidth=1, linecolor="black", showgrid=False)
    fig.update_yaxes(showticklabels=False, showline=True, mirror=True, linewidth=1, linecolor="black", showgrid=False)
    fig.update_layout(coloraxis={"colorscale": HISTOGRAM_COLORMAP, "cmin": vmin, "cmax": vmax})
    fig.update_layout(height=nrow * 300, width=ncol * 400)
    fig.update_layout(plot_bgcolor="rgba(0, 0, 0, 0)")
    fig.update_layout(hoverlabel=dict(bgcolor="white", font_size=16))
    fig.update_layout(bargap=0)

    fig = fig.to_dict()
    fig["layout"].update(shapes=[], annotations=[])

    for k, vstats in enumerate(all_vstats):
        row = (k // ncol) + 1
        col = (k % ncol) + 1
        subplot = k + 1
        plot_var(fig, draw, vstats, k, cbar_edges=cbar_edges, subplot=subplot, **kw)

    return fig


def plot_var(
    fig: "go.Figure", draw: "Draw", vstats: "VarStats", var: int, cbar_edges: np.ndarray, nbins=30, subplot=None
):
    values = draw.points[:, var].flatten()
    bin_range = vstats.p95_range
    sort_index = draw.get_argsort_indices(var)
    # bin_range = np.min(values), np.max(values)
    # showscale = (subplot == 1)
    showscale = True
    traces = _make_logp_histogram(
        values,
        draw.logp,
        nbins,
        bin_range,
        draw.weights,
        sort_index,
        cbar_edges=cbar_edges,
        showscale=showscale,
        subplot=subplot,
    )

    fig["data"].extend(traces)
    _decorate_histogram(vstats, fig, subplot=subplot)


def _decorate_histogram(vstats: "VarStats", fig: dict, subplot: int = 1):
    import plotly.graph_objects as go

    xaxis, yaxis = subplot_axis_names(subplot)
    l95, h95 = vstats.p95_range
    l68, h68 = vstats.p68_range
    # Shade things inside 1-sigma

    shading_rect = dict(
        type="rect",
        x0=l68,
        x1=h68,
        y0=0,
        y1=1,
        fillcolor="lightblue",
        opacity=0.5,
        layer="below",
        line={"width": 0},
        xref=xaxis,
        yref=f"{yaxis} domain",
    )
    fig["layout"]["shapes"].append(shading_rect)

    def marker(symbol, position, info_template):
        if position < l95:
            symbol, position, ha = "<" + symbol, l95, "left"
        elif position > h95:
            symbol, position, ha = ">" + symbol, h95, "right"
        else:
            symbol, position, ha = symbol, position, "center"
        new_marker = dict(
            xref=xaxis,
            yref=f"{yaxis} domain",
            x=position,
            y=0.95,
            text=symbol,
            showarrow=False,
            font=dict(color=ANNOTATION_COLOR),
            hovertext=info_template.format(position=position, label=vstats.label),
        )
        fig["layout"]["annotations"].append(new_marker)
        # pylab.axvline(v)

    marker("|", vstats.median, "{label}<br>median: {position}")
    marker("E", vstats.mean, "{label}<br>mean: {position}")
    marker("*", vstats.best, "{label}<br>best: {position}")

    label_annotation = dict(
        xref=f"{xaxis} domain",
        yref=f"{yaxis} domain",
        x=0.0,
        y=1.1,
        text=vstats.label,
        showarrow=False,
        name="label",
    )
    fig["layout"]["annotations"].append(label_annotation)


def _make_logp_histogram(values, logp, nbins, ci, weights, idx, cbar_edges, showscale=False, subplot=None):
    from numpy import ones_like, searchsorted, linspace, cumsum, diff, unique, argsort, array, hstack, exp

    if weights is None:
        weights = ones_like(logp)
    # use sorting index calculated during stats collection:
    values, weights, logp = values[idx], weights[idx], logp[idx]
    # open('/tmp/out','a').write("ci=%s, range=%s\n"
    #                           % (ci,(min(values),max(values))))
    edges = linspace(ci[0], ci[1], nbins + 1)
    bin_index = range(nbins)
    idx = searchsorted(values[1:-1], edges)

    bins = []  # marginalized maximum likelihood

    # H, xedges, yedges = np.histogram2d(values, -logp, bins=(edges, cbar_edges), weights=weights)
    # ybins = len(yedges) - 1
    # x = (xedges[:-1] + xedges[1:]) / 2.0
    # cbar_values = ((cbar_edges[:-1] + cbar_edges[1:]) / 2.0)[::-1]
    # x = x.repeat(ybins) # match length of y array
    # y = H.flatten(order="C")
    # color = np.tile(cbar_values, nbins)

    # # filter out empty y values (speeds up drawing, omitting empty rectangles):
    # y_nonzero = (y > 0)
    # # print(f"y_zeros: {y_nonzero}, {len([z for z in y_nonzero if z])} out of {len(y)}")
    # x = x[y_nonzero]
    # color = color[y_nonzero]
    # y = y[y_nonzero]
    # traces.append(go.Bar(x=x, y=y, showlegend=False, marker=dict(color=color, coloraxis='coloraxis', line=dict(width=0))))
    # # traces.append(go.Bar(x=x, y=y, showlegend=False, hoverinfo='skip', marker=dict(color=color, coloraxis='coloraxis', line=dict(width=0))))

    xscatt = []
    yscatt = []
    zscatt = []

    for i, s, e, xlo, xhi in zip(bin_index, idx[:-1], idx[1:], edges[:-1], edges[1:]):
        if s == e:
            continue

        pv = -logp[s:e]
        # weights for samples within interval
        pw = weights[s:e]
        # vertical colorbar top edges is the cumulative sum of the weights
        bin_height = pw.sum()

        # parameter interval endpoints
        x = array([xlo, xhi], "d")
        xav = (xlo + xhi) / 2
        # -logp values within interval, with sort index from low to high
        pv = -logp[s:e]
        pidx = argsort(pv)
        pv = pv[pidx]
        # weights for samples within interval, sorted
        pw = weights[s:e][pidx]
        # vertical colorbar top edges is the cumulative sum of the weights
        y_top = cumsum(pw)
        bin_height = y_top[-1]

        change_point = searchsorted(pv[1:-1], cbar_edges)
        tops = unique(hstack((change_point, len(pv) - 1)))
        # For plotly Bar plot, y values should be [size along y] (stacked)
        yy = np.diff(hstack((0, y_top[tops])))
        zz = pv[tops][:, None].flatten()
        xx = np.ones(tops.shape, dtype=float) * xav

        xscatt.extend(xx.tolist())
        yscatt.extend(yy.tolist())
        zscatt.extend(zz.tolist())

        # centerpoint, histogram height, maximum likelihood for each bin
        bin_max_likelihood = exp(cbar_edges[0] - pv[0])
        bins.append(((xlo + xhi) / 2, bin_height, bin_max_likelihood))

    xaxis, yaxis = subplot_axis_names(subplot)
    bar_trace = dict(
        type="bar",
        x=xscatt,
        y=yscatt,
        showlegend=False,
        hoverinfo="skip",
        marker=dict(color=zscatt, coloraxis="coloraxis", line=dict(width=0)),
        xaxis=xaxis,
        yaxis=yaxis,
    )
    # Check for broken distribution
    if not bins:
        return
    centers, height, maxlikelihood = array(bins).T
    # Normalize maximum likelihood plot so it contains the same area as the
    # histogram, unless it is really spikey, in which case make sure it has
    # about the same height as the histogram.
    maxlikelihood *= np.sum(height) / np.sum(maxlikelihood)
    hist_peak = np.max(height)
    ml_peak = np.max(maxlikelihood)
    if ml_peak > hist_peak * 1.3:
        maxlikelihood *= hist_peak * 1.3 / ml_peak

    scatter_trace = dict(
        type="scatter",
        x=centers,
        y=maxlikelihood,
        hoverinfo="skip",
        showlegend=False,
        line={"color": LINE_COLOR},
        xaxis=xaxis,
        yaxis=yaxis,
    )
    # return dict(bar_trace=bar_trace, scatter_trace=scatter_trace)
    return [bar_trace, scatter_trace]

    ## plot marginal gaussian approximation along with histogram
    # def G(x, mean, std):
    #    return np.exp(-((x-mean)/std)**2/2)/np.sqrt(2*np.pi*std**2)
    ## TODO: use weighted average for standard deviation
    # mean, std = np.average(values, weights=weights), np.std(values, ddof=1)
    # pdf = G(centers, mean, std)
    # pylab.plot(centers, pdf*np.sum(height)/np.sum(pdf), '-b')


def subplot_axis_names(subplot: int):
    if subplot == 1:
        return "x", "y"
    else:
        return f"x{subplot}", f"y{subplot}"



# Copyright (C) 2006-2011, University of Maryland
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/ or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# Author: James Krycka, Nikunj Patel

"""
This module implements the AppPanel class which creates the main panel on top
of the frame of the GUI for the Bumps application.
"""

# ==============================================================================

import os
import threading
from copy import deepcopy

import wx
import wx.aui

from .. import plugin
from ..cli import load_best, load_model
from ..dream import stats as dream_stats
from ..util import redirect_console
from . import signal
from .convergence_view import ConvergenceView
from .fit_dialog import show_fit_config
from .fit_thread import EVT_FIT_COMPLETE, EVT_FIT_PROGRESS, FitThread
from .log_view import LogView
from .parameter_view import ParameterView
from .plot_view import PlotView
from .summary_view import SummaryView
from .uncertainty_view import CorrelationView, ModelErrorView, TraceView, UncertaintyView
from .util import nice
from .utilities import get_bitmap, phoenix

# File selection strings.
MODEL_EXT = ".pickle"
MODEL_FILES = "Model files (*%s)|*%s" % (MODEL_EXT, MODEL_EXT)
PYTHON_FILES = "Script files (*.py)|*.py"
DATA_FILES = "Data files (*.dat)|*.dat"
TEXT_FILES = "Text files (*.txt)|*.txt"
ALL_FILES = "All files (*.*)|*"
PARS_FILES = "Parameter files (*.par)|*.par"

# Custom colors.
WINDOW_BKGD_COLOUR = "#ECE9D8"

# ==============================================================================


class AppPanel(wx.Panel):
    """
    This class builds the GUI for the application on a panel and attaches it
    to the frame.
    """

    def __init__(self, *args, **kw):
        # Create a panel on the frame.  This will be the only child panel of
        # the frame and it inherits its size from the frame which is useful
        # during resize operations (as it provides a minimal size to sizers).

        wx.Panel.__init__(self, *args, **kw)

        self.SetBackgroundColour("WHITE")

        # Modify the tool bar.
        frame = self.GetTopLevelParent()
        self.init_toolbar(frame)
        self.init_menubar(frame)

        # Reconfigure the status bar.
        self.init_statusbar(frame, [-34, -50, -16, -16])

        # Create the model views
        self.init_views()

        # Add data menu
        mb = frame.GetMenuBar()
        data_view = self.view["data"]
        if hasattr(data_view, "menu"):
            mb.Append(data_view.menu(), data_view.title)

        # Create a PubSub receiver.
        signal.connect(self.OnLogMessage, "log")
        signal.connect(self.OnModelNew, "model.new")
        signal.connect(self.OnModelChange, "model.update_structure")
        signal.connect(self.OnModelSetpar, "model.update_parameters")

        EVT_FIT_PROGRESS.Bind(self, wx.ID_ANY, wx.ID_ANY, self.OnFitProgress)
        EVT_FIT_COMPLETE.Bind(self, wx.ID_ANY, wx.ID_ANY, self.OnFitComplete)

        # _fit_abort and maybe _uncertainty_state are managed by fit_lock
        self.fit_lock = threading.Lock()
        self._fit_abort = False
        # self._uncertainty_state = None
        self.fit_thread = None
        self.fit_config = None

    def init_menubar(self, frame):
        """
        Adds items to the menu bar, menus, and menu options.
        The menu bar should already have a simple File menu and a Help menu.
        """
        mb = frame.GetMenuBar()

        file_menu_id = mb.FindMenu("File")
        file_menu = mb.GetMenu(file_menu_id)
        # help_menu = mb.GetMenu(mb.FindMenu("Help"))

        # Add items to the "File" menu (prepending them in reverse order).
        # Grey out items that are not currently implemented.
        file_menu.PrependSeparator()

        _item = file_menu.Prepend(wx.ID_ANY, "E&xport Results ...", "Save theory, data and parameters")
        frame.Bind(wx.EVT_MENU, self.OnFileExportResults, _item)

        _item = file_menu.Prepend(wx.ID_ANY, "&Reload", "Reload the existing model")
        frame.Bind(wx.EVT_MENU, self.OnFileReload, _item)

        _item = file_menu.Prepend(wx.ID_SAVEAS, "Save &As", "Save model as another name")
        frame.Bind(wx.EVT_MENU, self.OnFileSaveAs, _item)
        # file_menu.Enable(id=wx.ID_SAVEAS, enable=False)
        _item = file_menu.Prepend(wx.ID_SAVE, "&Save", "Save model")
        frame.Bind(wx.EVT_MENU, self.OnFileSave, _item)
        # file_menu.Enable(id=wx.ID_SAVE, enable=False)
        _item = file_menu.Prepend(wx.ID_OPEN, "&Open", "Open existing model")
        frame.Bind(wx.EVT_MENU, self.OnFileOpen, _item)
        # file_menu.Enable(id=wx.ID_OPEN, enable=False)
        _item = file_menu.Prepend(wx.ID_ANY, "Apply &Pararameters", "Apply parameters from .par file to loaded model")
        frame.Bind(wx.EVT_MENU, self.OnParsFileOpen, _item)
        _item = file_menu.Prepend(wx.ID_NEW, "&New", "Create new model")
        frame.Bind(wx.EVT_MENU, self.OnFileNew, _item)
        # file_menu.Enable(id=wx.ID_NEW, enable=False)

        # Add 'Fitting' menu to the menu bar and define its options.
        # Items are initially greyed out, but will be enabled after a script
        # is loaded.
        fit_menu = self.fit_menu = wx.Menu()

        _item = fit_menu.Append(wx.ID_ANY, "Start", "Start fitting operation")
        frame.Bind(wx.EVT_MENU, self.OnFitStart, _item)
        fit_menu.Enable(id=_item.GetId(), enable=False)
        self.fit_menu_start = _item

        _item = fit_menu.Append(wx.ID_ANY, "Stop", "Stop fitting operation")
        frame.Bind(wx.EVT_MENU, self.OnFitStop, _item)
        fit_menu.Enable(id=_item.GetId(), enable=False)
        self.fit_menu_stop = _item

        _item = fit_menu.Append(wx.ID_ANY, "&Options ...", "Edit fitting options")
        frame.Bind(wx.EVT_MENU, self.OnFitOptions, _item)
        self.fit_menu_options = _item

        # _item = fit_menu.Append(wx.ID_ANY,
        #                        "&Save ...",
        #                        "Save fit results")
        # frame.Bind(wx.EVT_MENU, self.OnFitSave, _item)
        # fit_menu.Enable(id=_item.GetId(), enable=False)
        self.fit_menu_options = _item

        mb.Append(fit_menu, "&Fitting")

    def init_toolbar(self, frame):
        """Populates the tool bar."""
        self.tb = frame.GetToolBar()

        script_bmp = get_bitmap("import_script.png", wx.BITMAP_TYPE_PNG)
        reload_bmp = get_bitmap("reload.png", wx.BITMAP_TYPE_PNG)
        start_bmp = get_bitmap("start_fit.png", wx.BITMAP_TYPE_PNG)
        stop_bmp = get_bitmap("stop_fit.png", wx.BITMAP_TYPE_PNG)

        _tool = self._add_tool(script_bmp, "Open model", "Load model from script")
        frame.Bind(wx.EVT_TOOL, self.OnFileOpen, _tool)
        _tool = self._add_tool(reload_bmp, "Reload model", "Reload model from script")
        frame.Bind(wx.EVT_TOOL, self.OnFileReload, _tool)
        # TODO: add reload

        self.tb.AddSeparator()

        _tool = self._add_tool(start_bmp, "Start Fit", "Start fitting operation")
        frame.Bind(wx.EVT_TOOL, self.OnFitStart, _tool)
        self.tb.EnableTool(_tool.GetId(), False)
        self.tb_start = _tool

        _tool = self._add_tool(stop_bmp, "Stop Fit", "Stop fitting operation")
        frame.Bind(wx.EVT_TOOL, self.OnFitStop, _tool)
        self.tb.EnableTool(_tool.GetId(), False)
        self.tb_stop = _tool

        self.tb.Realize()
        frame.SetToolBar(self.tb)

    def _add_tool(self, bitmap, label, help):
        if phoenix:
            return self.tb.AddTool(wx.ID_ANY, label, bitmap, shortHelp=help)
        else:
            return self.tb.AddSimpleTool(wx.ID_ANY, bitmap, label, help)

    def init_statusbar(self, frame, subbars):
        """Divides the status bar into multiple segments."""

        self.sb = frame.GetStatusBar()
        self.sb.SetFieldsCount(len(subbars))
        self.sb.SetStatusWidths(subbars)

    def init_views(self):
        # initial view
        self.aui = wx.aui.AuiNotebook(self)
        self.aui.Bind(wx.aui.EVT_AUINOTEBOOK_PAGE_CLOSE, self.OnViewTabClose)
        self.view_constructor = {
            "data": plugin.data_view(),
            "model": plugin.model_view(),
            "parameter": ParameterView,
            "summary": SummaryView,
            "log": LogView,
            "convergence": ConvergenceView,
            "uncertainty": UncertaintyView,
            "correlation": CorrelationView,
            "trace": TraceView,
            "error": ModelErrorView,
        }
        self.view_list = [
            "data",
            "model",
            "parameter",
            "summary",
            "log",
            "convergence",
            "uncertainty",
            "correlation",
            "trace",
            "error",
        ]
        self.view = {}
        for v in self.view_list:
            if self.view_constructor[v]:
                self.view[v] = self.view_constructor[v](self.aui, size=(600, 600))
                self.aui.AddPage(self.view[v], self.view_constructor[v].title)
        sizer = wx.BoxSizer(wx.VERTICAL)
        sizer.Add(self.aui, 1, wx.EXPAND)
        self.SetSizer(sizer)

    def show_view(self, tag):
        if self.view[tag].Parent == self.aui:
            self.aui.SetSelection(self.aui.GetPageIndex(self.view[tag]))
        else:
            self.view[tag].Raise()
            self.view[tag].SetFocus()

    def OnViewTabClose(self, evt):
        win = self.aui.GetPage(evt.GetSelection())
        # print "Closing tab",win.GetId()
        for k, w in self.view.items():
            if w == win:
                tag = k
                break
        else:
            raise RuntimeError("Lost track of view")
        # print "creating external frame"
        state = self.view[tag].get_state()
        constructor = self.view_constructor[tag]
        frame = wx.Frame(self, title=constructor.title, size=constructor.default_size)
        panel = constructor(frame)
        self.view[tag] = panel
        sizer = wx.BoxSizer(wx.VERTICAL)
        sizer.Add(panel, 1, wx.EXPAND)
        frame.SetSizer(sizer)
        frame.Bind(wx.EVT_CLOSE, self.OnViewFrameClose)
        frame.Show()
        panel.set_state(state)
        evt.Skip()

    def OnViewFrameClose(self, evt):
        win = evt.GetEventObject()
        # print "Closing frame",win.GetId()
        for k, w in self.view.items():
            if w.GetParent() == win:
                tag = k
                break
        else:
            raise RuntimeError("Lost track of view!")
        state = self.view[tag].get_state()
        constructor = self.view_constructor[tag]
        panel = constructor(self.aui)
        self.view[tag] = panel
        self.aui.AddPage(panel, constructor.title)
        panel.set_state(state)
        evt.Skip()

    # model viewer interface
    def OnLogMessage(self, message):
        for v in self.view.values():
            if hasattr(v, "log_message"):
                v.log_message(message)

    def OnModelNew(self, model):
        self.set_model(model)

    def OnModelChange(self, model):
        for v in self.view.values():
            if hasattr(v, "update_model"):
                v.update_model(model)

    def OnModelSetpar(self, model):
        for v in self.view.values():
            if hasattr(v, "update_parameters"):
                v.update_parameters(model)

    def OnFileNew(self, event):
        self.new_model()

    def OnFileOpen(self, event):
        # Load the script which will contain model definition and data.
        dlg = wx.FileDialog(
            self,
            message="Select File",
            # defaultDir=os.getcwd(),
            # defaultFile="",
            wildcard=(ALL_FILES),
            style=wx.FD_OPEN | wx.FD_CHANGE_DIR,
        )

        # Wait for user to close the dialog.
        status = dlg.ShowModal()
        path = dlg.GetPath()
        dlg.Destroy()

        # Process file if user clicked okay.
        if status == wx.ID_OK:
            self.load_model(path)

    def OnParsFileOpen(self, event):
        # Load a parameters (.pars) file to apply to current problem.
        dlg = wx.FileDialog(
            self,
            message="Select Parameters File",
            # defaultDir=os.getcwd(),
            # defaultFile="",
            wildcard=PARS_FILES,
            style=wx.FD_OPEN,
        )

        # Wait for user to close the dialog.
        status = dlg.ShowModal()
        path = dlg.GetPath()
        dlg.Destroy()

        # Process file if user clicked okay.
        if status == wx.ID_OK:
            self.apply_parameters(path)

    def OnFileReload(self, event):
        path = getattr(self, "_reload_path", self.model.path)
        self.load_model(path)

    def OnFileSave(self, event):
        if self.model is not None:
            # Force file extension to be MODEL_EXT
            self.model.path = os.path.splitext(self.model.path)[0] + MODEL_EXT
            self.save_model(self.model.path)
        else:
            self.OnFileSaveAs(event)

    def OnFileSaveAs(self, event):
        dlg = wx.FileDialog(
            self,
            message="Select File",
            defaultDir=os.getcwd(),
            defaultFile="",
            wildcard=(MODEL_FILES + "|" + ALL_FILES),
            style=wx.FD_SAVE | wx.FD_CHANGE_DIR | wx.FD_OVERWRITE_PROMPT,
        )
        # Wait for user to close the dialog.
        status = dlg.ShowModal()
        path = dlg.GetPath()
        dlg.Destroy()

        # Process file if user clicked okay.
        if status == wx.ID_OK:
            self.model.path = os.path.splitext(path)[0] + MODEL_EXT
            self.save_model(self.model.path)

    def OnFileExportResults(self, event):
        dlg = wx.DirDialog(self, message="Export results", defaultPath=os.getcwd(), style=wx.DD_DEFAULT_STYLE)
        # Wait for user to close the dialog.
        status = dlg.ShowModal()
        path = dlg.GetPath()
        dlg.Destroy()

        # Process file if user clicked okay.
        if status == wx.ID_OK:
            self.save_results(path)

    def set_fit_config(self, fit_config):
        self.fit_config = fit_config

    def OnFitOptions(self, event):
        # TODO: send fit_config to dialog rather than using options.FIT_CONFIG.
        # Wait until sasview wx is abandoned before making this change.
        show_fit_config(self)

    def OnFitStart(self, event):
        if self.fit_thread:
            self.sb.SetStatusText("Error: Fit already running")
            return
        # TODO: better access to model parameters
        if len(self.model.getp()) == 0:
            raise ValueError("Problem has no fittable parameters")

        # Start a new thread worker and give fit problem to the worker.
        fitclass = self.fit_config.selected_fitter
        options = self.fit_config.selected_values

        # Clear abort and uncertainty state
        self.abort = False
        self.uncertainty_state = None
        self.fit_thread = FitThread(
            win=self,
            abort_test=self._is_abort,
            problem=self.model,
            fitclass=fitclass,
            options=options,
            # Number of seconds between updates to the GUI, or 0 for no updates
            convergence_update=5,
            uncertainty_update=3600,
        )
        self.fit_thread.start()
        self.sb.SetStatusText("Fit status: Running", 3)

    def _is_abort(self):
        return self.abort

    @property
    def abort(self):
        with self.fit_lock:
            return self._fit_abort

    @abort.setter
    def abort(self, state):
        with self.fit_lock:
            self._fit_abort = state

    @property
    def uncertainty_state(self):
        """
        MCMC chain created by the DREAM optimizer.

        This is a thread-locked attribute maintained by the fitting thread.
        A deep copy of the attribute is made before setting so that state
        can continue to update within the fit thread while the user interface
        accesses the last state that was set.
        """
        with self.fit_lock:
            return self._uncertainty_state

    @uncertainty_state.setter
    def uncertainty_state(self, state):
        state_copy = deepcopy(state)
        with self.fit_lock:
            self._uncertainty_state = state_copy

    def OnFitStop(self, event):
        self.abort = True

    def OnFitComplete(self, event):
        if self.fit_thread is not None:
            self.fit_thread.join(1)  # 1 second timeout on join
            if self.fit_thread.is_alive():
                signal.log_message(message="fit thread failed to complete")
        self.fit_thread = None
        chisq = nice(2 * event.value / event.problem.dof)
        event.problem.setp(event.point)
        signal.update_parameters(model=event.problem)
        signal.log_message(message="done with chisq %g" % chisq)
        signal.log_message(message=event.info)
        self.sb.SetStatusText("Fit status: Complete", 3)
        beep()

    def OnFitSave(self, event):
        raise NotImplementedError()
        dlg = wx.FileDialog(
            self,
            message="Fit results",
            defaultDir=os.getcwd(),
            defaultFile="",
            wildcard=(MODEL_FILES + "|" + ALL_FILES),
            style=wx.FD_SAVE | wx.FD_CHANGE_DIR | wx.FD_OVERWRITE_PROMPT,
        )
        # Wait for user to close the dialog.
        status = dlg.ShowModal()
        path = dlg.GetPath()
        dlg.Destroy()

        # Process file if user clicked okay.
        if status == wx.ID_OK:
            self.save_fit(path)

    def OnFitProgress(self, event):
        self._fit_progress(event)
        return

        # Note: The following code was used to debugging plotting memory
        # and speed issues. It is here in case it might be useful again.
        if event.message == "uncertainty_final":
            event.message = "uncertainty_update"  # don't do model uncertainty
            n = 0
            import gc
            import time

            import psutil

            pid = os.getpid()
            proc = psutil.Process()
            t0 = time.perf_counter()
            signal.log_message("****** entering cycle ******")
            while True:
                self._fit_progress(event)
                t1 = time.perf_counter()
                msg = "========== cycle %d pid %d time %.3f s ===========" % (n, pid, t1 - t0)
                t0 = t1
                # signal.log_message(msg)
                # signal.log_message(f"memory {proc.memory_full_info()}")
                gc.collect()
                print(msg)
                print("memory", proc.memory_full_info())
                n += 1
                wx.Yield()
                # if n > 100: break

    def _fit_progress(self, event):
        if event.message == "progress":
            chisq = nice(2 * event.value / event.problem.dof)
            message = "step %5d chisq %g" % (event.step, chisq)
            signal.log_message(message=message)
        elif event.message == "improvement":
            event.problem.setp(event.point)
            event.problem.model_update()
            signal.update_parameters(model=event.problem)
        elif event.message == "convergence_update":
            self.view["convergence"].OnFitProgress(event)
        elif event.message in ("uncertainty_update", "uncertainty_final"):
            # TODO: revert to event.uncertainty_state if memory leak isn't fixed
            # Note: uncertainty_state is updated directly by the fit thread.
            # It's a copy protected by fit_lock so it should be self-consistent.
            state = self.uncertainty_state
            stats = dream_stats.var_stats(state.draw())
            self.console["state"] = self.uncertainty_state
            self.view["uncertainty"].fit_progress(event.problem, state, stats)
            self.view["correlation"].fit_progress(event.problem, state)
            self.view["trace"].fit_progress(event.problem, state)
            # TODO: send parameter stats to a view
            # It would be nice to have a "log scale" log of parameter history,
            # where the density of entries decreases exponentially over time.
            # So for example, the last 5, then 10 20 30 40 50, then 100, 200,
            # 300, 400, 500, ... Maybe have a slider so that you can drag it
            # forward and backward and see changes over time. Use box plots
            # to represent mode, median, mean, 68% and 95% intervals. Export
            # latest or all to CSV or text for pasting into word or excel.
            if event.message == "uncertainty_final":
                self.view["error"].fit_progress(event.problem, state)
                # Format the final variable nicely and show them on the console.
                # Note: only done at the end because GUI gets overwhelmed when
                # the text control has too much text.
                signal.log_message(dream_stats.format_vars(stats))
        else:
            raise ValueError("Unknown fit progress message " + event.message)

    def new_model(self):
        from ..plugin import new_model as gen

        self.set_model(gen())

    def load_model(self, path):
        self._reload_path = path
        model = load_model(path)
        signal.model_new(model=model)

    def apply_parameters(self, path):
        load_best(self.model, path)
        # TODO: remove this once LHS randomization is controlled from the command line. Ticket #51
        if hasattr(self.model, "undefined"):
            del self.model.undefined
        signal.update_parameters(model=self.model)

    def save_model(self, path):
        try:
            import dill

            dump = lambda obj, fid: dill.dump(obj, fid, recurse=True)
        except ImportError:
            from pickle import dump
        try:
            if hasattr(plugin, "save_json"):
                plugin.save_json(self.model, path)
            elif hasattr(self.model, "save_json"):
                self.model.save_json(path)
            with open(path, "wb") as fid:
                dump(self.model, fid)
        except Exception:
            import traceback

            signal.log_message(message=traceback.format_exc())

    def save_results(self, path):
        output_path = os.path.join(path, self.model.name)

        # Storage directory
        if not os.path.exists(path):
            os.mkdir(path)

        # Ask model to save its information
        self.model.save(output_path)

        # Save a snapshot of the model that can (hopefully) be reloaded
        self.save_model(output_path + MODEL_EXT)

        # Save the current state of the parameters
        with redirect_console(output_path + ".out"):
            self.model.show()
        pardata = "".join("%s %.15g\n" % (name, value) for name, value in zip(self.model.labels(), self.model.getp()))
        open(output_path + ".par", "wt").write(pardata)

        # Produce model plots
        self.model.plot(figfile=output_path)

        # Produce uncertainty plots
        if self.uncertainty_state is not None:
            with redirect_console(output_path + ".err"):
                self.uncertainty_state.show(figfile=output_path)
            self.uncertainty_state.save(output_path)

    def _add_measurement_type(self, type):
        """
        Add the panels needed to view a measurement of the given type.

        *type* is fitness.__class__, where fitness is the measurement cost function.
        """
        name = type.__name__
        if type not in self.data_tabs:
            tab = self.data_notebook.add_tab(type, name + " Data")
            constructor = getattr(type, "data_panel", PlotView)
            constructor(tab)
        if type not in self.model_notebook and hasattr(type, "model_panel"):
            tab = self.model_notebook.add_tab(type, name + " Model")
            type.model_panel(tab)

    def _view_problem(self, problem):
        """
        Set the model and data views to those necessary to display the problem.
        """
        # What types of measurements do we have?
        models = problem.models if hasattr(problem, "models") else [problem]
        types = set(p.fitness.__class__ for p in models)
        for p in types:
            self._add_measurement_type(p)

        # Show only the relevant views
        for p, tab in self.data_notebook.tabs():
            tab.Show(p in types)
        for p, tab in self.model_notebook.tabs():
            tab.Show(p in types)

    def set_model(self, model):
        # Inform the various tabs that the model they are viewing has changed.
        self.model = model

        # Point all of our views at the new model
        for v in self.view.values():
            if hasattr(v, "set_model"):
                v.set_model(model)
        self.console["model"] = model

        # Enable appropriate menu items.
        self.fit_menu.Enable(id=self.fit_menu_start.GetId(), enable=True)
        self.fit_menu.Enable(id=self.fit_menu_stop.GetId(), enable=True)
        self.fit_menu.Enable(id=self.fit_menu_options.GetId(), enable=True)

        # Enable appropriate toolbar items.
        self.tb.EnableTool(self.tb_start.GetId(), True)
        self.tb.EnableTool(self.tb_stop.GetId(), True)
        if hasattr(model, "broken_constraints"):
            signal.log_message(message="Unsatisfied constraints: \n\t%s" % (",\n\t".join(model.broken_constraints)))
        if hasattr(model, "path"):
            signal.log_message(message="loaded " + model.path)
            self.GetTopLevelParent().SetTitle("Bumps: %s" % model.name)
        else:
            signal.log_message(message="new model")
            self.GetTopLevelParent().SetTitle("Bumps")


SOUND = None


def beep():
    """
    Play fit completion sound.
    """
    wx.Bell()
    ## FIXME why doesn't sound work?
    # global SOUND
    # if SOUND is None:
    #    SOUND = wx.Sound(resource('done.wav'))
    # if SOUND.IsOk():
    #    SOUND.Play(wx.SOUND_ASYNC)



"""
Estimate model uncertainty from random sample.

MCMC uncertainty analysis gives the uncertainty on the model parameters
rather than the model itself.  For example, when fitting a line to a set
of data, the uncertainty on the slope and the intercept does not directly
give you the uncertainty in the expected value of *y* for a given value
of *x*.

The routines in bumps.errplot allow you to generate confidence intervals
on  the model using a random sample of MCMC parameters.  After calculating
the model *y* values for each sample, one can generate 68% and 95% contours
for a set of sampling points *x*.  This can apply even to models which
are not directly measured.  For example, in scattering inverse problems
the scattered intensity is the value measured, but the fitting parameters
describe the real space model that is being probed.  It is the uncertainty
in the real space model that is of primary interest.

Since bumps knows only the probability of seeing the measured value given
the input parameters, it is up to the model itself to calculate and display
the confidence intervals on the model and the expected values for the data
points.  This is done using the :mod:`bumps.plugin` architecture, so
application writers can provide the appropriate functions for their data
types.  Eventually this capability will move to the model definition so
that different types of models can be processed in the same fit.

For a completed MCMC run, four steps are required:

#. reload the fitting problem and the MCMC state
#. select a set of sample points
#. evaluate model confidence intervals from sample points
#. show model confidence intervals

:func:`reload_errors` performs steps 1, 2 and 3, returning *errs*.
If the fitting problem and the MCMC state are already loaded, then use
:func:`calc_errors_from_state` to perform steps 2 and 3, returning *errs*.
If alternative sampling is desired, then use :func:`calc_errors` on a
given set of points to perform step 3, returning *errs*.  Once *errs* has
been calculated and returned by one of these methods, call
:func:`show_errors` to perform step 4.
"""

__all__ = ["reload_errors", "calc_errors_from_state", "calc_errors", "show_errors"]
import os
import traceback
import logging

import numpy as np

from .dream.state import load_state
from . import plugin
from .cli import load_model, load_best


def reload_errors(model, store, nshown=50, random=True):
    """
    Reload the MCMC state and compute the model confidence intervals.

    The loaded error data is a sample from the fit space according to the
    fit parameter uncertainty.  This is a subset of the samples returned
    by the DREAM MCMC sampling process.

    *model* is the name of the model python file

    *store* is the name of the store directory containing the dream results

    *nshown* and *random* are as for :func:`calc_errors_from_state`.

    Returns *errs* for :func:`show_errors`.
    """
    problem = load_model(model)
    load_best(problem, os.path.join(store, problem.name + ".par"))
    state = load_state(os.path.join(store, problem.name))
    state.mark_outliers()
    return calc_errors_from_state(problem, state, nshown=nshown, random=random)


def error_points_from_state(state, nshown=50, random=True, portion=1.0):
    """
    Return a set of points from the state for calculating errors.

    *nshown* is the number of samples to include from the state.

    *random* is True if the samples are randomly selected, or False if
    the most recent samples should be used.  Use random if you have
    poor mixing (i.e., the parameters tend to stay fixed from generation
    to generation), but not random if your burn-in was too short, and
    you want to select from the end.

    Returns *points* for :func:`calc_errors`.
    """

    points, _logp = state.sample(portion=portion)
    if points.shape[0] < nshown:
        nshown = points.shape[0]
    # randomize the draw; skip the last point since state.keep_best() put
    # the best point at the end.
    if random:
        points = points[np.random.permutation(len(points) - 1)]
    return points[-nshown:-1]


def calc_errors_from_state(problem, state, nshown=50, random=True, portion=1.0):
    """
    Compute confidence regions for a problem from the
    Align the sample profiles and compute the residual difference from
    the measured data for a set of points returned from DREAM.

    Returns *errs* for :func:`show_errors`.
    """

    points = error_points_from_state(state, nshown=nshown, random=random, portion=portion)
    return calc_errors(problem, points)


def calc_errors(problem, points):
    """
    Align the sample profiles and compute the residual difference from the
    measured data for a set of points.

    The return value is arbitrary.  It is passed to the :func:`show_errors`
    plugin for the application.
    Returns *errs* for :func:`show_errors`.
    """
    original = problem.getp()
    try:
        ret = plugin.calc_errors(problem, points)
    except Exception:
        info = ["error calculating distribution on model", traceback.format_exc()]
        logging.error("\n".join(info))
        ret = None
    finally:
        problem.setp(original)
    return ret


def show_errors(errs, fig=None):
    """
    Display the confidence regions returned by :func:`calc_errors`.

    The content of *errs* depends on the active plugin.
    """
    return plugin.show_errors(errs, fig=fig)



"""
Signals changes to the model that need to be reflected in the views.

In practice, the main window is the only listener, and it forwards the
messages to the appropriate views.
"""

import wx
import wx.py.dispatcher
from wx.py.dispatcher import send

# export the connect function; do it this way so that linters don't complain
# about unused imports
connect = wx.py.dispatcher.connect


def model_new(model):
    """
    Inform all views that a new model is available.
    """
    wx.CallAfter(send, "model.new", model=model)


def update_model(model, dirty=True):
    """
    Inform all views that the model structure has changed.  This calls
    model.model_reset() to reset the fit parameters and constraints.
    """
    model.model_reset()  #
    if dirty:
        model.model_update()
    wx.CallAfter(send, "model.update_structure", model=model)


_DELAYED_SIGNAL = {}


def update_parameters(model, delay=100):
    """
    Inform all views that the model has changed.  Note that if the model
    is changing rapidly, then the signal will be delayed for a time.  This
    calls model.model_update() to let the model know that it needs to be
    recalculated.
    """
    # signaller is responsible for marking the model as needing recalculation
    model.model_update()
    # TODO: potential race condition
    # Future call mat be occurring at the time that restart is triggered.
    # Not sure we can do anything about it from outside wx...
    signal = _DELAYED_SIGNAL.get(model, None)
    if signal is not None:
        # signal is already active, so delay it some more
        signal.Restart(delay)
    else:
        # activate a new signal, and call when back at GUI loop
        def _send_signal():
            # print "sending update parameters",model
            del _DELAYED_SIGNAL[model]
            wx.CallAfter(send, "model.update_parameters", model=model)

        _DELAYED_SIGNAL[model] = wx.CallLater(delay, _send_signal)


def log_message(message):
    wx.CallAfter(send, "log", message=message)
